
import numpy as np

def daylen(DOY, XLAT):
    # Calculation of declination of sun (Eqn. 16). Amplitude= +/-23.45
    # deg. Minimum = DOY 355 (DEC 21), maximum = DOY 172.5 (JUN 21/22).
    # Sun angles.  SOC limited for latitudes above polar circles.
    # Calculate daylength, sunrise and sunset (Eqn. 17)

    RAD=np.pi/180.0

    DEC = -23.45 * np.cos(2.0*np.pi*(DOY+10.0)/365.0)

    SOC = np.tan(RAD*DEC) * np.tan(RAD*XLAT)
    SOC = min(max(SOC,-1.0),1.0)
    # SOC=alt

    DAYL = 12.0 + 24.0*np.arcsin(SOC)/np.pi
    SNUP = 12.0 - DAYL/2.0
    SNDN = 12.0 + DAYL/2.0

    return DAYL, DEC, SNDN, SNUP



# Generated with SMOP  0.41
from libsmop import *
# Ratio_Kb.m

    
@function
def Ratio_Kb(Kin=None,A=None,lat=None,d=None,t=None,Atr=None,*args,**kwargs):
    varargin = Ratio_Kb.varargin
    nargin = Ratio_Kb.nargin

    # function is used to estimate incoming shortwave diffuse radiation under
# clear sky conditions with inputs A(alititude, m), lat(latitude, degrees),
# d (days, Jan 1 =1), t (time, 24 hours)
    
    m=opt_m(A,lat,d,t)
# Ratio_Kb.m:8
    Kb=Kin / ((1 + (multiply(0.3,(1 - Atr ** m)))) / (Atr ** m))
# Ratio_Kb.m:10


# Generated with SMOP  0.41
from libsmop import *
# opt_m.m

    
@function
def opt_m(A=None,lat=None,d=None,t=None,*args,**kwargs):
    varargin = opt_m.varargin
    nargin = opt_m.nargin

    #optical air mass number used for input into the equations which estimate
#incoming shortwave radiation under clear sky conditions
    
    Pa=atm_P(A)
# opt_m.m:6
    zen=solar_zenith(lat,d,t)
# opt_m.m:8
    m=Pa / (multiply(101.3,cosd(zen)))
# opt_m.m:10
    # written January/07 by Natasha Kenny


import numpy as np
from .sunonsurface_2018a import sunonsurface_2018a
# import matplotlib.pyplot as plt


def gvf_2018a(wallsun, walls, buildings, scale, shadow, first, second, dirwalls, Tg, Tgwall, Ta, emis_grid, ewall,
              alb_grid, SBC, albedo_b, rows, cols, Twater, lc_grid, landcover):
    azimuthA = np.arange(5, 359, 20)  # Search directions for Ground View Factors (GVF)

    #### Ground View Factors ####
    gvfLup = np.zeros((rows, cols))
    gvfalb = np.zeros((rows, cols))
    gvfalbnosh = np.zeros((rows, cols))
    gvfLupE = np.zeros((rows, cols))
    gvfLupS = np.zeros((rows, cols))
    gvfLupW = np.zeros((rows, cols))
    gvfLupN = np.zeros((rows, cols))
    gvfalbE = np.zeros((rows, cols))
    gvfalbS = np.zeros((rows, cols))
    gvfalbW = np.zeros((rows, cols))
    gvfalbN = np.zeros((rows, cols))
    gvfalbnoshE = np.zeros((rows, cols))
    gvfalbnoshS = np.zeros((rows, cols))
    gvfalbnoshW = np.zeros((rows, cols))
    gvfalbnoshN = np.zeros((rows, cols))
    gvfSum = np.zeros((rows, cols))

    #  sunwall=wallinsun_2015a(buildings,azimuth(i),shadow,psi(i),dirwalls,walls);
    sunwall = (wallsun / walls * buildings) == 1  # new as from 2015a

    for j in np.arange(0, azimuthA.__len__()):
        _, gvfLupi, gvfalbi, gvfalbnoshi, gvf2 = sunonsurface_2018a(azimuthA[j], scale, buildings, shadow, sunwall,
                                                                    first,
                                                                    second, dirwalls * np.pi / 180, walls, Tg, Tgwall,
                                                                    Ta,
                                                                    emis_grid, ewall, alb_grid, SBC, albedo_b, Twater,
                                                                    lc_grid, landcover)

        gvfLup = gvfLup + gvfLupi
        gvfalb = gvfalb + gvfalbi
        gvfalbnosh = gvfalbnosh + gvfalbnoshi
        gvfSum = gvfSum + gvf2

        if (azimuthA[j] >= 0) and (azimuthA[j] < 180):
            gvfLupE = gvfLupE + gvfLupi
            gvfalbE = gvfalbE + gvfalbi
            gvfalbnoshE = gvfalbnoshE + gvfalbnoshi

        if (azimuthA[j] >= 90) and (azimuthA[j] < 270):
            gvfLupS = gvfLupS + gvfLupi
            gvfalbS = gvfalbS + gvfalbi
            gvfalbnoshS = gvfalbnoshS + gvfalbnoshi

        if (azimuthA[j] >= 180) and (azimuthA[j] < 360):
            gvfLupW = gvfLupW + gvfLupi
            gvfalbW = gvfalbW + gvfalbi
            gvfalbnoshW = gvfalbnoshW + gvfalbnoshi

        if (azimuthA[j] >= 270) or (azimuthA[j] < 90):
            gvfLupN = gvfLupN + gvfLupi
            gvfalbN = gvfalbN + gvfalbi
            gvfalbnoshN = gvfalbnoshN + gvfalbnoshi

    gvfLup = gvfLup / azimuthA.__len__() + SBC * emis_grid * (Ta + 273.15) ** 4
    gvfalb = gvfalb / azimuthA.__len__()
    gvfalbnosh = gvfalbnosh / azimuthA.__len__()

    gvfLupE = gvfLupE / (azimuthA.__len__() / 2) + SBC * emis_grid * (Ta + 273.15) ** 4
    gvfLupS = gvfLupS / (azimuthA.__len__() / 2) + SBC * emis_grid * (Ta + 273.15) ** 4
    gvfLupW = gvfLupW / (azimuthA.__len__() / 2) + SBC * emis_grid * (Ta + 273.15) ** 4
    gvfLupN = gvfLupN / (azimuthA.__len__() / 2) + SBC * emis_grid * (Ta + 273.15) ** 4

    gvfalbE = gvfalbE / (azimuthA.__len__() / 2)
    gvfalbS = gvfalbS / (azimuthA.__len__() / 2)
    gvfalbW = gvfalbW / (azimuthA.__len__() / 2)
    gvfalbN = gvfalbN / (azimuthA.__len__() / 2)

    gvfalbnoshE = gvfalbnoshE / (azimuthA.__len__() / 2)
    gvfalbnoshS = gvfalbnoshS / (azimuthA.__len__() / 2)
    gvfalbnoshW = gvfalbnoshW / (azimuthA.__len__() / 2)
    gvfalbnoshN = gvfalbnoshN / (azimuthA.__len__() / 2)

    gvfNorm = gvfSum / (azimuthA.__len__())
    gvfNorm[buildings == 0] = 1
    
    return gvfLup, gvfalb, gvfalbnosh, gvfLupE, gvfalbE, gvfalbnoshE, gvfLupS, gvfalbS, gvfalbnoshS, gvfLupW, gvfalbW, gvfalbnoshW, gvfLupN, gvfalbN, gvfalbnoshN, gvfSum, gvfNorm


# Generated with SMOP  0.41
from libsmop import *
# solar_dec.m

    
@function
def solar_dec(d=None,*args,**kwargs):
    varargin = solar_dec.varargin
    nargin = solar_dec.nargin

    # written to calcuate the solar declination with ranges from +23.45 at
# summer solstice to -23.45 at the winter solstice.  d is the calendar day
# with d = 1 being January 1st.
    
    #sin_dec = 0.39785 .* sind(278.97 + 0.9856 .* d + 1.9165 .* sind(356.6 +
#0.9856 .* d)); from Campbell and norman pg. 168
    
    #dec = asind(sin_dec);
    
    dec=dot(- 23.4,cosd(dot(360,(d + 10)) / 365))
# solar_dec.m:13
    
    # written January, 2007 by Natasha Kenny


import json
import logging
from types import SimpleNamespace
from typing import Any, Dict, List, Optional

import numpy as np

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# handle for QGIS which does not have matplotlib by default?
try:
    from matplotlib import pyplot as plt

    PLT = True
except ImportError:
    PLT = False

from ... import common
from ...class_configs import EnvironData, RasterData, ShadowMatrices, SolweigConfig, SvfData, TgMaps, WallsData
from . import PET_calculations
from . import Solweig_2025a_calc_forprocessing as so
from . import UTCI_calculations as utci
from .CirclePlotBar import PolarBarPlot
from .wallsAsNetCDF import walls_as_netcdf


def dict_to_namespace(d):
    """Recursively convert dicts to SimpleNamespace."""
    if isinstance(d, dict):
        return SimpleNamespace(**{k: dict_to_namespace(v) for k, v in d.items()})
    elif isinstance(d, list):
        return [dict_to_namespace(i) for i in d]
    else:
        return d


class SolweigRun:
    """Class to run the SOLWEIG algorithm with given configuration."""

    config: SolweigConfig
    progress: Optional[Any]
    iters_total: Optional[int]
    iters_count: int = 0
    poi_names: List[Any] = []
    poi_pixel_xys: Optional[np.ndarray]
    poi_results = []
    woi_names: List[Any] = []
    woi_pixel_xys: Optional[np.ndarray]
    woi_results = []
    raster_data: RasterData
    location: Dict[str, float]
    svf_data: SvfData
    environ_data: EnvironData
    tg_maps: TgMaps
    shadow_mats: ShadowMatrices
    walls_data: WallsData

    def __init__(
        self,
        config: SolweigConfig,
        params_json_path: str,
        amax_local_window_m: int = 100,
        amax_local_perc: float = 99.9,
    ):
        """Initialize the SOLWEIG runner with configuration and parameters."""
        logger.info("Starting SOLWEIG setup")
        self.config = config
        self.config.validate()
        # Progress tracking settings
        self.progress = None
        self.iters_total = None
        self.iters_count = 0
        self.proceed = True
        # Initialize POI data
        self.poi_names = []
        self.poi_pixel_xys = None
        self.poi_results = []
        # Initialize WOI data
        self.woi_names = []
        self.woi_pixel_xys = None
        self.woi_results = []
        # Load parameters from JSON file
        params_path = common.check_path(params_json_path)
        try:
            with open(params_path) as f:
                params_dict = json.load(f)
                self.params = dict_to_namespace(params_dict)
        except Exception as e:
            raise RuntimeError(f"Failed to load parameters from {params_json_path}: {e}")
        # Initialize SVF and Raster data
        self.svf_data = SvfData(self.config)
        self.raster_data = RasterData(
            self.config,
            self.params,
            self.svf_data,
            amax_local_window_m,
            amax_local_perc,
        )
        # Location data
        left_x = self.raster_data.trf_arr[0]
        top_y = self.raster_data.trf_arr[3]
        lng, lat = common.xy_to_lnglat(self.raster_data.crs_wkt, left_x, top_y)
        alt = float(np.nanmedian(self.raster_data.dsm))
        if alt < 0:
            alt = 3
        self.location = {"longitude": lng, "latitude": lat, "altitude": alt}
        # weather data
        if self.config.use_epw_file:
            self.environ_data = self.load_epw_weather()
            logger.info("Weather data loaded from EPW file")
        else:
            self.environ_data = self.load_met_weather(header_rows=1, delim=" ")
            logger.info("Weather data loaded from MET file")
        # POIs check
        if self.config.poi_path:
            self.load_poi_data()
            logger.info("POI data loaded from %s", self.config.poi_path)
        # Import shadow matrices (Anisotropic sky)
        self.shadow_mats = ShadowMatrices(self.config, self.params, self.svf_data)
        logger.info("Shadow matrices initialized")
        # % Ts parameterisation maps
        self.tg_maps = TgMaps(
            self.config.use_landcover,
            self.params,
            self.raster_data,
        )
        logger.info("TgMaps initialized")
        # Import data for wall temperature parameterization
        # Use wall of interest
        if self.config.woi_path:
            self.load_woi_data()
            logger.info("WOI data loaded from %s", self.config.woi_path)
        self.walls_data = WallsData(
            self.config,
            self.params,
            self.raster_data,
            self.environ_data,
            self.tg_maps,
        )
        logger.info("WallsData initialized")

    def test_hook(self) -> None:
        """Test hook for testing loaded init state."""
        pass

    def prep_progress(self, num: int) -> None:
        """Prepare progress for environment."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def iter_progress(self) -> bool:
        """Iterate progress ."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def load_epw_weather(self) -> EnvironData:
        """Load weather data from an EPW file."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def load_met_weather(self, header_rows: int = 1, delim: str = " ") -> EnvironData:
        """Load weather data from a MET file."""
        met_path_str = str(common.check_path(self.config.met_path))
        met_data = np.loadtxt(met_path_str, skiprows=header_rows, delimiter=delim, dtype=np.float32)
        return EnvironData(
            self.config,
            self.params,
            YYYY=met_data[:, 0],
            DOY=met_data[:, 1],
            hours=met_data[:, 2],
            minu=met_data[:, 3],
            Ta=met_data[:, 11],
            RH=met_data[:, 10],
            radG=met_data[:, 14],
            radD=met_data[:, 21],
            radI=met_data[:, 22],
            P=met_data[:, 12],
            Ws=met_data[:, 9],
            location=self.location,
            UTC=self.config.utc,
        )

    def load_poi_data(self) -> None:
        """Load point of interest (POI) data from a file."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def save_poi_results(self) -> None:
        """Save results for points of interest (POIs) to files."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def load_woi_data(self) -> None:
        """Load wall of interest (WOI) data from a file."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def save_woi_results(self) -> None:
        """Save results for walls of interest (WOIs) to files."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def hemispheric_image(self):
        """
        Calculate patch characteristics for points of interest (POIs).
        This method is vectorized for efficiency as it processes all POIs simultaneously.
        """
        n_patches = self.shadow_mats.shmat.shape[2]
        n_pois = self.poi_pixel_xys.shape[0]
        patch_characteristics = np.zeros((n_patches, n_pois), dtype=np.float32)

        # Get POI indices as integer arrays
        poi_y = self.poi_pixel_xys[:, 2].astype(int)
        poi_x = self.poi_pixel_xys[:, 1].astype(int)

        for idy in range(n_patches):
            # Precompute masks for this patch
            temp_sky = (self.shadow_mats.shmat[:, :, idy] == 1) & (self.shadow_mats.vegshmat[:, :, idy] == 1)
            temp_vegsh = (self.shadow_mats.vegshmat[:, :, idy] == 0) | (self.shadow_mats.vbshvegshmat[:, :, idy] == 0)
            temp_vbsh = (1 - self.shadow_mats.shmat[:, :, idy]) * self.shadow_mats.vbshvegshmat[:, :, idy]
            temp_sh = temp_vbsh == 1

            if self.config.use_wall_scheme:
                temp_sh_w = temp_sh * self.walls_data.voxelMaps[:, :, idy]
                temp_sh_roof = temp_sh * (self.walls_data.voxelMaps[:, :, idy] == 0)
            else:
                temp_sh_w = None
                temp_sh_roof = None

            # Gather mask values for all POIs at once
            sky_vals = temp_sky[poi_y, poi_x]
            veg_vals = temp_vegsh[poi_y, poi_x]
            sh_vals = temp_sh[poi_y, poi_x]

            if self.config.use_wall_scheme:
                sh_w_vals = temp_sh_w[poi_y, poi_x]
                sh_roof_vals = temp_sh_roof[poi_y, poi_x]

            # Assign patch characteristics in vectorized way
            patch_characteristics[idy, sky_vals] = 1.8
            patch_characteristics[idy, ~sky_vals & veg_vals] = 2.5
            if self.config.use_wall_scheme:
                patch_characteristics[idy, ~sky_vals & ~veg_vals & sh_vals & sh_w_vals] = 4.5
                patch_characteristics[idy, ~sky_vals & ~veg_vals & sh_vals & ~sh_w_vals & sh_roof_vals] = 4.5
            else:
                patch_characteristics[idy, ~sky_vals & ~veg_vals & sh_vals] = 4.5

        return patch_characteristics

    def calc_solweig(
        self,
        iter: int,
        elvis: float,
        first: float,
        second: float,
        firstdaytime: float,
        timeadd: float,
        timestepdec: float,
        posture,
    ):
        """
        Calculate SOLWEIG results for a given iteration.
        Separated from the main run method so that it can be overridden by subclasses.
        Over time we can simplify the function signature by passing consolidated classes to solweig calc methods.
        """
        return so.Solweig_2025a_calc(
            iter,
            self.raster_data.dsm,
            self.raster_data.scale,
            self.raster_data.rows,
            self.raster_data.cols,
            self.svf_data.svf,
            self.svf_data.svf_north,
            self.svf_data.svf_west,
            self.svf_data.svf_east,
            self.svf_data.svf_south,
            self.svf_data.svf_veg,
            self.svf_data.svf_veg_north,
            self.svf_data.svf_veg_east,
            self.svf_data.svf_veg_south,
            self.svf_data.svf_veg_west,
            self.svf_data.svf_veg_blocks_bldg_sh,
            self.svf_data.svf_veg_blocks_bldg_sh_east,
            self.svf_data.svf_veg_blocks_bldg_sh_south,
            self.svf_data.svf_veg_blocks_bldg_sh_west,
            self.svf_data.svf_veg_blocks_bldg_sh_north,
            self.raster_data.cdsm,
            self.raster_data.tdsm,
            self.params.Albedo.Effective.Value.Walls,
            self.params.Tmrt_params.Value.absK,
            self.params.Tmrt_params.Value.absL,
            self.params.Emissivity.Value.Walls,
            posture.Fside,
            posture.Fup,
            posture.Fcyl,
            self.environ_data.altitude[iter],
            self.environ_data.azimuth[iter],
            self.environ_data.zen[iter],
            self.environ_data.jday[iter],
            self.config.use_veg_dem,
            self.config.only_global,
            self.raster_data.buildings,
            self.location,
            self.environ_data.psi[iter],
            self.config.use_landcover,
            self.raster_data.lcgrid,
            self.environ_data.dectime[iter],
            self.environ_data.altmax[iter],
            self.raster_data.wallaspect,
            self.raster_data.wallheight,
            int(self.config.person_cylinder),  # expects int though should work either way
            elvis,
            self.environ_data.Ta[iter],
            self.environ_data.RH[iter],
            self.environ_data.radG[iter],
            self.environ_data.radD[iter],
            self.environ_data.radI[iter],
            self.environ_data.P[iter],
            self.raster_data.amaxvalue,
            self.raster_data.bush,
            self.environ_data.Twater[iter],
            self.tg_maps.TgK,
            self.tg_maps.Tstart,
            self.tg_maps.alb_grid,
            self.tg_maps.emis_grid,
            self.tg_maps.TgK_wall,
            self.tg_maps.Tstart_wall,
            self.tg_maps.TmaxLST,
            self.tg_maps.TmaxLST_wall,
            first,
            second,
            self.svf_data.svfalfa,
            self.raster_data.svfbuveg,
            firstdaytime,
            timeadd,
            timestepdec,
            self.tg_maps.Tgmap1,
            self.tg_maps.Tgmap1E,
            self.tg_maps.Tgmap1S,
            self.tg_maps.Tgmap1W,
            self.tg_maps.Tgmap1N,
            self.environ_data.CI[iter],
            self.tg_maps.TgOut1,
            self.shadow_mats.diffsh,
            self.shadow_mats.shmat,
            self.shadow_mats.vegshmat,
            self.shadow_mats.vbshvegshmat,
            int(self.config.use_aniso),  # expects int though should work either way
            self.shadow_mats.asvf,
            self.shadow_mats.patch_option,
            self.walls_data.voxelMaps,
            self.walls_data.voxelTable,
            self.environ_data.Ws[iter],
            self.config.use_wall_scheme,
            self.walls_data.timeStep,
            self.shadow_mats.steradians,
            self.walls_data.walls_scheme,
            self.walls_data.dirwalls_scheme,
        )

    def run(self) -> None:
        # Posture settings
        if self.params.Tmrt_params.Value.posture == "Standing":
            posture = self.params.Posture.Standing.Value
        else:
            posture = self.params.Posture.Sitting.Value
        # Radiative surface influence
        first = np.round(posture.height)
        if first == 0.0:
            first = 1.0
        second = np.round(posture.height * 20.0)
        # Save hemispheric image
        if self.config.use_aniso and self.poi_pixel_xys is not None:
            patch_characteristics = self.hemispheric_image()
            logger.info("Hemispheric image calculated for POIs")
        # Initialisation of time related variables
        if self.environ_data.Ta.__len__() == 1:
            timestepdec = 0
        else:
            timestepdec = self.environ_data.dectime[1] - self.environ_data.dectime[0]
        timeadd = 0.0
        firstdaytime = 1.0
        # Initiate array for I0 values plotting
        if np.unique(self.environ_data.DOY).shape[0] > 1:
            unique_days = np.unique(self.environ_data.DOY)
            first_unique_day = self.environ_data.DOY[unique_days[0] == self.environ_data.DOY]
            I0_array = np.zeros_like(first_unique_day, dtype=np.float32)
        else:
            first_unique_day = self.environ_data.DOY.copy()
            I0_array = np.zeros_like(self.environ_data.DOY, dtype=np.float32)
        # For Tmrt plot
        tmrt_agg = np.zeros((self.raster_data.rows, self.raster_data.cols), dtype=np.float32)
        # Number of iterations
        num = len(self.environ_data.Ta)
        # Prepare progress tracking
        self.prep_progress(num)
        logger.info("Progress tracking prepared for %d iterations", num)
        elvis = 0.0
        #
        for i in range(num):
            self.proceed = self.iter_progress()
            if not self.proceed:
                break
            self.iters_count += 1
            # Run the SOLWEIG calculations
            (
                Tmrt,
                Kdown,
                Kup,
                Ldown,
                Lup,
                Tg,
                ea,
                esky,
                I0,
                CI,
                shadow,
                firstdaytime,
                timestepdec,
                timeadd,
                self.tg_maps.Tgmap1,
                self.tg_maps.Tgmap1E,
                self.tg_maps.Tgmap1S,
                self.tg_maps.Tgmap1W,
                self.tg_maps.Tgmap1N,
                Keast,
                Ksouth,
                Kwest,
                Knorth,
                Least,
                Lsouth,
                Lwest,
                Lnorth,
                KsideI,
                self.tg_maps.TgOut1,
                TgOut,
                radIout,
                radDout,
                Lside,
                Lsky_patch_characteristics,
                CI_Tg,
                CI_TgG,
                KsideD,
                dRad,
                Kside,
                self.shadow_mats.steradians,
                voxelTable,
            ) = self.calc_solweig(
                i,
                elvis,
                first,
                second,
                firstdaytime,
                timeadd,
                timestepdec,
                posture,
            )

            # Aggregate Tmrt
            # Guard against NaN and Inf - replace non-finite with avg if available
            if (~np.isfinite(Tmrt)).any() and self.iters_count > 1:
                logger.warning("Tmrt contains non-finite values, replacing with preceding average.")
                tmrt_avg = tmrt_agg / self.iters_count
                tmrt_agg = np.where(np.isfinite(Tmrt), tmrt_agg + Tmrt, tmrt_avg)
            elif (~np.isfinite(tmrt_agg)).any():
                raise ValueError("Tmrt aggregation contains non-finite values.")
            else:
                tmrt_agg = tmrt_agg + Tmrt

            # Save I0 for I0 vs. Kdown output plot to check if UTC is off
            if i < first_unique_day.shape[0]:
                I0_array[i] = I0
            elif i == first_unique_day.shape[0] and PLT is True:
                # Output I0 vs. Kglobal plot
                radG_for_plot = self.environ_data.radG[first_unique_day[0] == self.environ_data.DOY]
                dectime_for_plot = self.environ_data.dectime[first_unique_day[0] == self.environ_data.DOY]
                fig, ax = plt.subplots()
                ax.plot(dectime_for_plot, I0_array, label="I0")
                ax.plot(dectime_for_plot, radG_for_plot, label="Kglobal")
                ax.set_ylabel("Shortwave radiation [$Wm^{-2}$]")
                ax.set_xlabel("Decimal time")
                ax.set_title("UTC" + str(self.config.utc))
                ax.legend()
                fig.savefig(self.config.output_dir + "/metCheck.png", dpi=150)

            if self.environ_data.altitude[i] > 0:
                w = "D"
            else:
                w = "N"

            if self.environ_data.hours[i] < 10:
                XH = "0"
            else:
                XH = ""

            if self.environ_data.minu[i] < 10:
                XM = "0"
            else:
                XM = ""

            if self.poi_pixel_xys is not None:
                for n in range(0, self.poi_pixel_xys.shape[0]):
                    idx, row_idx, col_idx = self.poi_pixel_xys[n]
                    row_idx = int(row_idx)
                    col_idx = int(col_idx)
                    result_row = {
                        "poi_idx": idx,
                        "col_idx": col_idx,
                        "row_idx": row_idx,
                        "yyyy": self.environ_data.YYYY[i],
                        "id": self.environ_data.jday[i],
                        "it": self.environ_data.hours[i],
                        "imin": self.environ_data.minu[i],
                        "dectime": self.environ_data.dectime[i],
                        "altitude": self.environ_data.altitude[i],
                        "azimuth": self.environ_data.azimuth[i],
                        "kdir": radIout,
                        "kdiff": radDout,
                        "kglobal": self.environ_data.radG[i],
                        "kdown": Kdown[row_idx, col_idx],
                        "kup": Kup[row_idx, col_idx],
                        "keast": Keast[row_idx, col_idx],
                        "ksouth": Ksouth[row_idx, col_idx],
                        "kwest": Kwest[row_idx, col_idx],
                        "knorth": Knorth[row_idx, col_idx],
                        "ldown": Ldown[row_idx, col_idx],
                        "lup": Lup[row_idx, col_idx],
                        "least": Least[row_idx, col_idx],
                        "lsouth": Lsouth[row_idx, col_idx],
                        "lwest": Lwest[row_idx, col_idx],
                        "lnorth": Lnorth[row_idx, col_idx],
                        "Ta": self.environ_data.Ta[i],
                        "Tg": TgOut[row_idx, col_idx],
                        "RH": self.environ_data.RH[i],
                        "Esky": esky,
                        "Tmrt": Tmrt[row_idx, col_idx],
                        "I0": I0,
                        "CI": CI,
                        "Shadow": shadow[row_idx, col_idx],
                        "SVF_b": self.svf_data.svf[row_idx, col_idx],
                        "SVF_bv": self.raster_data.svfbuveg[row_idx, col_idx],
                        "KsideI": KsideI[row_idx, col_idx],
                    }
                    # Recalculating wind speed based on powerlaw
                    WsPET = (1.1 / self.params.Wind_Height.Value.magl) ** 0.2 * self.environ_data.Ws[i]
                    WsUTCI = (10.0 / self.params.Wind_Height.Value.magl) ** 0.2 * self.environ_data.Ws[i]
                    resultPET = PET_calculations._PET(
                        self.environ_data.Ta[i],
                        self.environ_data.RH[i],
                        Tmrt[row_idx, col_idx],
                        WsPET,
                        self.params.PET_settings.Value.Weight,
                        self.params.PET_settings.Value.Age,
                        self.params.PET_settings.Value.Height,
                        self.params.PET_settings.Value.Activity,
                        self.params.PET_settings.Value.clo,
                        self.params.PET_settings.Value.Sex,
                    )
                    result_row["PET"] = resultPET
                    resultUTCI = utci.utci_calculator(
                        self.environ_data.Ta[i], self.environ_data.RH[i], Tmrt[row_idx, col_idx], WsUTCI
                    )
                    result_row["UTCI"] = resultUTCI
                    result_row["CI_Tg"] = CI_Tg
                    result_row["CI_TgG"] = CI_TgG
                    result_row["KsideD"] = KsideD[row_idx, col_idx]
                    result_row["Lside"] = Lside[row_idx, col_idx]
                    result_row["diffDown"] = dRad[row_idx, col_idx]
                    result_row["Kside"] = Kside[row_idx, col_idx]
                    self.poi_results.append(result_row)

            if self.config.use_wall_scheme and self.woi_pixel_xys is not None:
                for n in range(0, self.woi_pixel_xys.shape[0]):
                    idx, row_idx, col_idx = self.woi_pixel_xys[n]
                    row_idx = int(row_idx)
                    col_idx = int(col_idx)

                    temp_wall = voxelTable.loc[
                        ((voxelTable["ypos"] == row_idx) & (voxelTable["xpos"] == col_idx)), "wallTemperature"
                    ].to_numpy()
                    K_in = voxelTable.loc[
                        ((voxelTable["ypos"] == row_idx) & (voxelTable["xpos"] == col_idx)), "K_in"
                    ].to_numpy()
                    L_in = voxelTable.loc[
                        ((voxelTable["ypos"] == row_idx) & (voxelTable["xpos"] == col_idx)), "L_in"
                    ].to_numpy()
                    wallShade = voxelTable.loc[
                        ((voxelTable["ypos"] == row_idx) & (voxelTable["xpos"] == col_idx)), "wallShade"
                    ].to_numpy()

                    result_row = {
                        "woi_idx": idx,
                        "woi_name": self.woi_names[idx],
                        "yyyy": self.environ_data.YYYY[i],
                        "id": self.environ_data.jday[i],
                        "it": self.environ_data.hours[i],
                        "imin": self.environ_data.minu[i],
                        "dectime": self.environ_data.dectime[i],
                        "Ta": self.environ_data.Ta[i],
                        "SVF": self.svf_data.svf[row_idx, col_idx],
                        "Ts": temp_wall,
                        "Kin": K_in,
                        "Lin": L_in,
                        "shade": wallShade,
                        "pixel_x": col_idx,
                        "pixel_y": row_idx,
                    }
                    self.woi_results.append(result_row)

                if self.config.wall_netcdf:
                    netcdf_output = self.config.output_dir + "/walls.nc"
                    walls_as_netcdf(
                        voxelTable,
                        self.raster_data.rows,
                        self.raster_data.cols,
                        self.walls_data.met_for_xarray,
                        i,
                        self.raster_data.dsm,
                        self.config.dsm_path,
                        netcdf_output,
                    )

            time_code = (
                str(int(self.environ_data.YYYY[i]))
                + "_"
                + str(int(self.environ_data.DOY[i]))
                + "_"
                + XH
                + str(int(self.environ_data.hours[i]))
                + XM
                + str(int(self.environ_data.minu[i]))
                + w
            )

            if self.config.output_tmrt:
                common.save_raster(
                    self.config.output_dir + "/Tmrt_" + time_code + ".tif",
                    Tmrt,
                    self.raster_data.trf_arr,
                    self.raster_data.crs_wkt,
                    self.raster_data.nd_val,
                    coerce_f64_to_f32=True,
                )
            if self.config.output_kup:
                common.save_raster(
                    self.config.output_dir + "/Kup_" + time_code + ".tif",
                    Kup,
                    self.raster_data.trf_arr,
                    self.raster_data.crs_wkt,
                    self.raster_data.nd_val,
                    coerce_f64_to_f32=True,
                )
            if self.config.output_kdown:
                common.save_raster(
                    self.config.output_dir + "/Kdown_" + time_code + ".tif",
                    Kdown,
                    self.raster_data.trf_arr,
                    self.raster_data.crs_wkt,
                    self.raster_data.nd_val,
                    coerce_f64_to_f32=True,
                )
            if self.config.output_lup:
                common.save_raster(
                    self.config.output_dir + "/Lup_" + time_code + ".tif",
                    Lup,
                    self.raster_data.trf_arr,
                    self.raster_data.crs_wkt,
                    self.raster_data.nd_val,
                    coerce_f64_to_f32=True,
                )
            if self.config.output_ldown:
                common.save_raster(
                    self.config.output_dir + "/Ldown_" + time_code + ".tif",
                    Ldown,
                    self.raster_data.trf_arr,
                    self.raster_data.crs_wkt,
                    self.raster_data.nd_val,
                    coerce_f64_to_f32=True,
                )
            if self.config.output_sh:
                common.save_raster(
                    self.config.output_dir + "/Shadow_" + time_code + ".tif",
                    shadow,
                    self.raster_data.trf_arr,
                    self.raster_data.crs_wkt,
                    self.raster_data.nd_val,
                    coerce_f64_to_f32=True,
                )
            if self.config.output_kdiff:
                common.save_raster(
                    self.config.output_dir + "/Kdiff_" + time_code + ".tif",
                    dRad,
                    self.raster_data.trf_arr,
                    self.raster_data.crs_wkt,
                    self.raster_data.nd_val,
                    coerce_f64_to_f32=True,
                )

            # Sky view image of patches
            if (
                i == 0
                and PLT is True
                and self.config.plot_poi_patches
                and self.config.use_aniso
                and self.poi_pixel_xys is not None
            ):
                for k in range(self.poi_pixel_xys.shape[0]):
                    Lsky_patch_characteristics[:, 2] = patch_characteristics[:, k]
                    skyviewimage_out = self.config.output_dir + "/POI_" + str(self.poi_names[k]) + ".png"
                    PolarBarPlot(
                        Lsky_patch_characteristics,
                        self.environ_data.altitude[i],
                        self.environ_data.azimuth[i],
                        "Hemisphere partitioning",
                        skyviewimage_out,
                        0,
                        5,
                        0,
                    )

        # Abort if loop was broken
        if not self.proceed:
            return

        # Save POI results
        if self.poi_results:
            self.save_poi_results()

        # Save WOI results
        if self.woi_results:
            self.save_woi_results()

        # Save Tree Planter results
        if self.config.output_tree_planter:
            pos = 1 if self.params.Tmrt_params.Value.posture == "Standing" else 0

            settingsHeader = [
                "UTC",
                "posture",
                "onlyglobal",
                "landcover",
                "anisotropic",
                "cylinder",
                "albedo_walls",
                "albedo_ground",
                "emissivity_walls",
                "emissivity_ground",
                "absK",
                "absL",
                "elevation",
                "patch_option",
            ]
            settingsFmt = (
                "%i",
                "%i",
                "%i",
                "%i",
                "%i",
                "%i",
                "%1.2f",
                "%1.2f",
                "%1.2f",
                "%1.2f",
                "%1.2f",
                "%1.2f",
                "%1.2f",
                "%i",
            )
            settingsData = np.array(
                [
                    [
                        int(self.config.utc),
                        pos,
                        self.config.only_global,
                        self.config.use_landcover,
                        self.config.use_aniso,
                        self.config.person_cylinder,
                        self.params.Albedo.Effective.Value.Walls,
                        self.params.Albedo.Effective.Value.Cobble_stone_2014a,
                        self.params.Emissivity.Value.Walls,
                        self.params.Emissivity.Value.Cobble_stone_2014a,
                        self.params.Tmrt_params.Value.absK,
                        self.params.Tmrt_params.Value.absL,
                        self.location["altitude"],
                        self.shadow_mats.patch_option,
                    ]
                ],
                dtype=np.float32,
            )
            np.savetxt(
                self.config.output_dir + "/treeplantersettings.txt",
                settingsData,
                fmt=settingsFmt,
                header=", ".join(settingsHeader),
                delimiter=" ",
            )

        # Save average Tmrt raster
        if self.iters_count > 0:
            tmrt_avg = tmrt_agg / self.iters_count
            common.save_raster(
                self.config.output_dir + "/Tmrt_average.tif",
                tmrt_avg,
                self.raster_data.trf_arr,
                self.raster_data.crs_wkt,
                self.raster_data.nd_val,
                coerce_f64_to_f32=True,
            )



from pathlib import Path
from typing import Optional

import numpy as np

from umep.class_configs import SolweigConfig
from umep.functions.SOLWEIGpython.solweig_runner_core import SolweigRunCore

PROJECT_ROOT = Path(__file__).resolve().parents[1]
CONFIG_PATH = PROJECT_ROOT / "demos/data/athens/configsolweig.ini"
PARAMS_PATH = PROJECT_ROOT / "demos/data/athens/parametersforsolweig.json"


def _prepare_temp_config(tmp_path: Path) -> Path:
    """Return a config file that writes outputs into a temp dir."""
    config = SolweigConfig()
    config.from_file(str(CONFIG_PATH))

    output_dir = tmp_path / "output"
    working_dir = tmp_path / "working"

    config.output_dir = str(output_dir)
    config.working_dir = str(working_dir)

    path_overrides = {
        "dsm_path": PROJECT_ROOT / "demos/data/athens/DSM.tif",
        "cdsm_path": PROJECT_ROOT / "temp/athens/CDSM.tif",
        "dem_path": PROJECT_ROOT / "demos/data/athens/DEM.tif",
        "wh_path": PROJECT_ROOT / "temp/athens/walls/wall_hts.tif",
        "wa_path": PROJECT_ROOT / "temp/athens/walls/wall_aspects.tif",
        "svf_path": PROJECT_ROOT / "temp/athens/svf/svfs.zip",
        "aniso_path": PROJECT_ROOT / "temp/athens/svf/shadowmats.npz",
        "epw_path": PROJECT_ROOT / "demos/data/athens/athens_2023.epw",
    }
    for attr, path in path_overrides.items():
        setattr(config, attr, str(path))

    tmp_config = tmp_path / "configsolweig.ini"
    config.to_file(str(tmp_config))
    return tmp_config


def _assert_float32(array: Optional[np.ndarray], name: str):
    if array is not None:
        assert array.dtype == np.float32, f"{name} dtype was {array.dtype}, expected float32"


def test_athens_runner_rasters_are_float32(tmp_path):
    tmp_config = _prepare_temp_config(tmp_path)
    runner = SolweigRunCore(str(tmp_config), str(PARAMS_PATH))
    raster_data = runner.raster_data

    _assert_float32(raster_data.dsm, "dsm")
    _assert_float32(raster_data.wallheight, "wallheight")
    _assert_float32(raster_data.wallaspect, "wallaspect")
    _assert_float32(raster_data.dem, "dem")
    _assert_float32(raster_data.cdsm, "cdsm")
    _assert_float32(raster_data.tdsm, "tdsm")
    _assert_float32(raster_data.bush, "bush")
    _assert_float32(raster_data.svfbuveg, "svfbuveg")
    _assert_float32(raster_data.buildings, "buildings")

    runner.test_hook()



"""
/***************************************************************************
 ProcessingUMEP
                                 A QGIS plugin
 UMEP for processing toolbox
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2020-04-02
        copyright            : (C) 2020 by Fredrik Lindberg
        email                : fredrikl@gvc.gu.se
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

"SOLWEIG (v2022a) is a model which can be used to estimate spatial variations of 3D radiation fluxes and "
"mean radiant temperature (Tmrt) in complex urban settings. The SOLWEIG model follows the same "
"approach commonly adopted to observe Tmrt, with shortwave and longwave radiation fluxes from  "
"six directions being individually calculated to derive Tmrt. The model requires a limited number "
"of inputs, such as direct, diffuse and global shortwave radiation, air temperature, relative "
"humidity, urban geometry and geographical information (latitude, longitude and elevation). "
"Additional vegetation and ground cover information can also be used to imporove the estimation of Tmrt.\n"
"\n"
"Tools to generate sky view factors, wall height and aspect etc. is available in the pre-processing past in UMEP\n"
"\n"
"------------\n"
"\n"
"Full manual available via the <b>Help</b>-button."
"https://umep-docs.readthedocs.io/en/latest/processor/Outdoor%20Thermal%20Comfort%20SOLWEIG.html"
"""

__author__ = "Fredrik Lindberg"
__date__ = "2020-04-02"
__copyright__ = "(C) 2020 by Fredrik Lindberg"

# This will get replaced with a git SHA1 when you do a git archive

__revision__ = "$Format:%H$"

import random
import string
import zipfile
from pathlib import Path
from shutil import rmtree

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pyproj
from pvlib.iotools import read_epw
from rasterio.transform import Affine, rowcol, xy
from tqdm import tqdm

from umep import common
from umep.functions.SOLWEIGpython import PET_calculations as p
from umep.functions.SOLWEIGpython import Solweig_2022a_calc_forprocessing as so
from umep.functions.SOLWEIGpython import UTCI_calculations as utci
from umep.functions.SOLWEIGpython import WriteMetadataSOLWEIG_old
from umep.util.SEBESOLWEIGCommonFiles.clearnessindex_2013b import clearnessindex_2013b
from umep.util.SEBESOLWEIGCommonFiles.Solweig_v2015_metdata_noload import (
    Solweig_2015a_metdata_noload,
)


def generate_solweig(
    dsm_path: str,
    wall_ht_path: str,
    wall_aspect_path: str,
    svf_path: str,
    epw_path: str,
    bbox: list[int, int, int, int],
    out_dir: str,
    start_date_Ymd: str,  # %Y-%m-%d"
    end_date_Ymd: str,  # %Y-%m-%d"
    hours: list[int] = list(range(1, 25)),
    veg_dsm_path: str | None = None,
    pois_gdf: gpd.GeoDataFrame | None = None,
    trans_veg: float = 3,
    trunk_zone_ht_perc: float = 0.25,
    leaf_start: int = 97,
    leaf_end: int = 300,
    conif_trees: bool = False,
    albedo_bldg: float = 0.2,
    albedo_ground: float = 0.15,
    emmisiv_bldg: float = 0.9,
    emmisiv_ground: float = 0.95,
    body_shortwave_absorp: float = 0.7,
    body_longwave_absorp: float = 0.95,
    estimate_radiation_from_global=False,
):
    as_cylinder = 0
    standing = True
    if standing is True:
        Fside = 0.22
        Fup = 0.06
        height = 1.1  # METRES!
        Fcyl = 0.28
    else:
        Fside = 0.166666
        Fup = 0.166666
        height = 0.75  # METRES!
        Fcyl = 0.2
    # for PET
    age = 35
    activity = 80
    clothing = 0.9
    weight = 75
    pers_ht = 1.8  # METRES!  # different from above height param
    sex = "male"
    wind_sensor_ht = 10
    utc = 0

    # veg transmissivity as percentage
    if not trans_veg >= 0 and trans_veg <= 100:
        raise ValueError("Vegetation transmissivity should be a number between 0 and 100")
    trans_veg = trans_veg / 100.0

    out_path = Path(out_dir)
    out_path.mkdir(parents=True, exist_ok=True)
    out_path_str = str(out_path)

    Path.mkdir(out_path / "shadows", parents=True, exist_ok=True)
    Path.mkdir(out_path / "Tmrt", parents=True, exist_ok=True)

    dsm, dsm_transf, dsm_crs, _dsm_nd = common.load_raster(dsm_path, bbox, coerce_f64_to_f32=True)
    dsm_trf_affine = Affine.from_gdal(*dsm_transf)
    dsm_scale = 1 / dsm_transf[1]
    dsm_height, dsm_width = dsm.shape  # y rows by x cols
    # y is flipped - so return max for lower row
    minx, miny = xy(dsm_trf_affine, dsm.shape[0], 0)
    # Define the source and target CRS
    source_crs = pyproj.CRS(dsm_crs)
    target_crs = pyproj.CRS(4326)  # WGS 84
    # Create a transformer object
    transformer = pyproj.Transformer.from_crs(source_crs, target_crs, always_xy=True)
    # Perform the transformation
    lon, lat = transformer.transform(minx, miny)

    alt = np.median(dsm)
    if alt < 3:
        alt = 3

    trunkfile = 0
    trunkratio = 0

    if veg_dsm_path is not None:
        usevegdem = 1
        veg_dsm, veg_dsm_transf, veg_dsm_crs, _veg_dsm_nd = common.load_raster(
            veg_dsm_path, bbox, coerce_f64_to_f32=True
        )
        veg_dsm_height, veg_dsm_width = veg_dsm.shape
        if not (veg_dsm_width == dsm_width) & (veg_dsm_height == dsm_height):
            raise ValueError("Error in Vegetation Canopy DSM: All rasters must be of same extent and resolution")
        trunkratio = trunk_zone_ht_perc / 100.0
        veg_dsm_2 = veg_dsm * trunkratio
        veg_dsm_2_height, veg_dsm_2_width = veg_dsm_2.shape
        if not (veg_dsm_2_width == dsm_width) & (veg_dsm_2_height == dsm_height):
            raise ValueError("Error in Trunk Zone DSM: All rasters must be of same extent and resolution")
        veg_dsm_2_path = None
    else:
        usevegdem = 0
        veg_dsm = np.zeros([dsm_height, dsm_width])
        veg_dsm_2 = np.zeros([dsm_height, dsm_width])
        veg_dsm_2_path = None

    # Land cover
    filePath_lc = None
    landcover = 0

    # DEM
    demforbuild = 0

    # SVFs
    temp_dir_name = "temp-" + "".join(random.choice(string.ascii_uppercase) for _ in range(8))
    temp_dir = out_path_str + "/" + temp_dir_name
    zip = zipfile.ZipFile(svf_path, "r")
    zip.extractall(temp_dir)
    zip.close()

    svf, _, _, _ = common.load_raster(temp_dir + "/svf.tif", bbox, coerce_f64_to_f32=True)
    svfN, _, _, _ = common.load_raster(temp_dir + "/svfN.tif", bbox, coerce_f64_to_f32=True)
    svfS, _, _, _ = common.load_raster(temp_dir + "/svfS.tif", bbox, coerce_f64_to_f32=True)
    svfE, _, _, _ = common.load_raster(temp_dir + "/svfE.tif", bbox, coerce_f64_to_f32=True)
    svfW, _, _, _ = common.load_raster(temp_dir + "/svfW.tif", bbox, coerce_f64_to_f32=True)

    if usevegdem == 1:
        svfveg, _, _, _ = common.load_raster(temp_dir + "/svfveg.tif", bbox, coerce_f64_to_f32=True)
        svfNveg, _, _, _ = common.load_raster(temp_dir + "/svfNveg.tif", bbox, coerce_f64_to_f32=True)
        svfSveg, _, _, _ = common.load_raster(temp_dir + "/svfSveg.tif", bbox, coerce_f64_to_f32=True)
        svfEveg, _, _, _ = common.load_raster(temp_dir + "/svfEveg.tif", bbox, coerce_f64_to_f32=True)
        svfWveg, _, _, _ = common.load_raster(temp_dir + "/svfWveg.tif", bbox, coerce_f64_to_f32=True)
        svfaveg, _, _, _ = common.load_raster(temp_dir + "/svfaveg.tif", bbox, coerce_f64_to_f32=True)
        svfNaveg, _, _, _ = common.load_raster(temp_dir + "/svfNaveg.tif", bbox, coerce_f64_to_f32=True)
        svfSaveg, _, _, _ = common.load_raster(temp_dir + "/svfSaveg.tif", bbox, coerce_f64_to_f32=True)
        svfEaveg, _, _, _ = common.load_raster(temp_dir + "/svfEaveg.tif", bbox, coerce_f64_to_f32=True)
        svfWaveg, _, _, _ = common.load_raster(temp_dir + "/svfWaveg.tif", bbox, coerce_f64_to_f32=True)
    else:
        svfveg = np.ones((dsm_height, dsm_width))
        svfNveg = np.ones((dsm_height, dsm_width))
        svfSveg = np.ones((dsm_height, dsm_width))
        svfEveg = np.ones((dsm_height, dsm_width))
        svfWveg = np.ones((dsm_height, dsm_width))
        svfaveg = np.ones((dsm_height, dsm_width))
        svfNaveg = np.ones((dsm_height, dsm_width))
        svfSaveg = np.ones((dsm_height, dsm_width))
        svfEaveg = np.ones((dsm_height, dsm_width))
        svfWaveg = np.ones((dsm_height, dsm_width))

    svf_dsm_height, svf_dsm_width = svfveg.shape
    if not (svf_dsm_width == dsm_width) & (svf_dsm_height == dsm_height):
        raise ValueError("Error in SVF: All rasters must be of same extent and resolution")
    tmp = svf + svfveg - 1.0
    tmp[tmp < 0.0] = 0.0
    svfalfa = np.arcsin(np.exp(np.log(1.0 - tmp) / 2.0))

    wh_rast, wh_transf, wh_crs, _wh_nd = common.load_raster(wall_ht_path, bbox, coerce_f64_to_f32=True)
    wh_height, wh_width = wh_rast.shape
    if not (wh_width == dsm_width) & (wh_height == dsm_height):
        raise ValueError("Error in Wall height raster: All rasters must be of same extent and resolution")
    wa_rast, wa_transf, wa_crs, _wa_nd = common.load_raster(wall_aspect_path, bbox, coerce_f64_to_f32=True)
    wa_height, wa_width = wa_rast.shape
    if not (wa_width == dsm_width) & (wa_height == dsm_height):
        raise ValueError("Error in Wall aspect raster: All rasters must be of same extent and resolution")

    # Metdata
    metfileexist = 1
    epw_df, epw_info = read_epw(epw_path)
    # Filter by date range
    filtered_df = epw_df.loc[start_date_Ymd:end_date_Ymd]
    # Filter by hours
    filtered_df = filtered_df[filtered_df.index.hour.isin(hours)]
    # raise if empty
    if len(filtered_df) == 0:
        raise ValueError("No EPW dates intersect start and end dates and / or hours.")
    umep_df = pd.DataFrame(
        {
            "iy": filtered_df.index.year,
            "id": filtered_df.index.dayofyear,
            "it": filtered_df.index.hour,
            "imin": filtered_df.index.minute,
            "Q": -999,
            "QH": -999,
            "QE": -999,
            "Qs": -999,
            "Qf": -999,
            "Wind": filtered_df["wind_speed"],
            "RH": filtered_df["relative_humidity"],
            "Tair": filtered_df["temp_air"],
            "pres": filtered_df["atmospheric_pressure"].astype(float),  # Pascal, ensure float
            "rain": -999,
            "Kdown": filtered_df["ghi"],
            "snow": filtered_df["snow_depth"],
            "ldown": filtered_df["ghi_infrared"],
            "fcld": filtered_df["total_sky_cover"],
            "wuh": filtered_df["precipitable_water"],
            "xsmd": -999,
            "lai_hr": -999,
            "Kdiff": filtered_df["dhi"],
            "Kdir": filtered_df["dni"],
            "Wdir": filtered_df["wind_direction"],
        }
    )
    umep_df_filt = umep_df[(umep_df["Kdown"] < 0) & (umep_df["Kdown"] > 1300)]
    if len(umep_df_filt):
        print(umep_df_filt.head())
        raise ValueError(
            "Error: Kdown - beyond what is expected",
        )

    # use -999 for NaN to mesh with UMEP
    umep_df = umep_df.fillna(-999)

    print("Calculating sun positions for each time step")
    met_data = umep_df.to_numpy()
    location = {"longitude": lon, "latitude": lat, "altitude": alt}
    YYYY, altitude, azimuth, zen, jday, leafon, dectime, altmax = Solweig_2015a_metdata_noload(met_data, location, utc)

    # Creating vectors from meteorological input
    DOY = umep_df.loc[:, "id"].values
    hours = umep_df.loc[:, "it"].values
    minu = umep_df.loc[:, "imin"].values
    Ta = umep_df.loc[:, "Tair"].values
    RH = umep_df.loc[:, "RH"].values
    radG = umep_df.loc[:, "Kdown"].values
    radD = umep_df.loc[:, "Kdiff"].values
    radI = umep_df.loc[:, "Kdir"].values
    umep_df.loc[:, "pres"] = umep_df.loc[:, "pres"] / 100.0  # convert from Pa to hPa
    P = umep_df.loc[:, "pres"].values
    Ws = umep_df.loc[:, "Wind"].values

    # %Radiative surface influence, Rule of thumb by Schmid et al. (1990).
    first = np.round(height)
    if first == 0.0:
        first = 1.0
    second = np.round(height * 10.0)  # NOTE: using 10 instead of 20

    if usevegdem == 1:
        # Conifer or deciduous
        if conif_trees is True:
            leafon = np.ones((1, DOY.shape[0]))
        else:
            leafon = np.zeros((1, DOY.shape[0]))
            if leaf_start > leaf_end:
                leaf_bool = (leaf_start < DOY) | (leaf_end > DOY)
            else:
                leaf_bool = (leaf_start < DOY) & (leaf_end > DOY)
            leafon[0, leaf_bool] = 1

        # % Vegetation transmittivity of shortwave radiation
        psi = leafon * trans_veg
        psi[leafon == 0] = 0.5
        # amaxvalue
        vegmax = veg_dsm.max()
        amaxvalue = dsm.max() - dsm.min()
        amaxvalue = np.maximum(amaxvalue, vegmax)

        # Elevation vegdsms if buildingDEM includes ground heights
        veg_dsm = veg_dsm + dsm
        veg_dsm[veg_dsm == dsm] = 0
        veg_dsm_2 = veg_dsm_2 + dsm
        veg_dsm_2[veg_dsm_2 == dsm] = 0

        # % Bush separation
        bush = np.logical_not(veg_dsm_2 * veg_dsm) * veg_dsm

        svfbuveg = svf - (1.0 - svfveg) * (1.0 - trans_veg)  # % major bug fixed 20141203
    else:
        psi = leafon * 0.0 + 1.0
        svfbuveg = svf
        bush = np.zeros([dsm_height, dsm_width])
        amaxvalue = 0

    # %Initialization of maps
    Knight = np.zeros((dsm_height, dsm_width))
    Tgmap1 = np.zeros((dsm_height, dsm_width))
    Tgmap1E = np.zeros((dsm_height, dsm_width))
    Tgmap1S = np.zeros((dsm_height, dsm_width))
    Tgmap1W = np.zeros((dsm_height, dsm_width))
    Tgmap1N = np.zeros((dsm_height, dsm_width))

    buildings = dsm.copy()
    buildings[buildings < 2.0] = 1.0
    buildings[buildings >= 2.0] = 0.0

    # Import shadow matrices (Anisotropic sky)
    anisotropic_sky = 0
    diffsh = None
    shmat = None
    vegshmat = None
    vbshvegshmat = None
    asvf = None
    patch_option = 0

    # % Ts parameterisation maps
    TgK = Knight + 0.37
    Tstart = Knight - 3.41
    alb_grid = Knight + albedo_ground
    emis_grid = Knight + emmisiv_ground
    TgK_wall = 0.37
    Tstart_wall = -3.41
    TmaxLST = 15.0
    TmaxLST_wall = 15.0

    # Initialisation of time related variables
    if Ta.__len__() == 1:
        timestepdec = 0
    else:
        timestepdec = dectime[1] - dectime[0]
    timeadd = 0.0
    firstdaytime = 1.0

    WriteMetadataSOLWEIG_old.writeRunInfo(
        out_path_str,
        dsm_path,
        source_crs,
        usevegdem,
        veg_dsm_path,
        trunkfile,
        veg_dsm_2_path,
        lat,
        lon,
        utc,
        landcover,
        filePath_lc,
        metfileexist,
        epw_path,
        met_data,
        body_shortwave_absorp,
        body_longwave_absorp,
        albedo_bldg,
        albedo_ground,
        emmisiv_bldg,
        emmisiv_ground,
        estimate_radiation_from_global,
        trunkratio,
        trans_veg,
        dsm_height,
        dsm_width,
        int(standing),
        0,  # elvis
        as_cylinder,
        demforbuild,
        anisotropic_sky,
        None,
        None,
    )

    print("Writing settings for this model run to specified output folder (Filename: RunInfoSOLWEIG_YYYY_DOY_HHMM.txt)")

    #  If metfile starts at night
    CI = 1.0

    # Main function
    print("Executing main model")

    tmrtplot = np.zeros((dsm_height, dsm_width))
    TgOut1 = np.zeros((dsm_height, dsm_width))

    if pois_gdf is not None:
        poi_results = []

    # Initiate array for I0 values
    if np.unique(DOY).shape[0] > 1:
        unique_days = np.unique(DOY)
        first_unique_day = DOY[unique_days[0] == DOY]
        I0_array = np.zeros(first_unique_day.shape[0])
    else:
        first_unique_day = DOY.copy()
        I0_array = np.zeros(DOY.shape[0])

    for i in tqdm(np.arange(0, Ta.__len__())):
        # Nocturnal cloudfraction from Offerle et al. 2003
        if (dectime[i] - np.floor(dectime[i])) == 0:
            daylines = np.where(np.floor(dectime) == dectime[i])
            if daylines.__len__() > 1:
                alt = altitude[0][daylines]
                alt2 = np.where(alt > 1)
                rise = alt2[0][0]
                [_, CI, _, _, _] = clearnessindex_2013b(
                    zen[0, i + rise + 1],
                    jday[0, i + rise + 1],
                    Ta[i + rise + 1],
                    RH[i + rise + 1] / 100.0,
                    radG[i + rise + 1],
                    location,
                    P[i + rise + 1],
                )  # i+rise+1 to match matlab code. correct?
                if (CI > 1.0) or (np.inf == CI):
                    CI = 1.0
            else:
                CI = 1.0

        # radI[i] = radI[i]/np.sin(altitude[0][i] * np.pi/180)

        (
            Tmrt,
            Kdown,
            Kup,
            Ldown,
            Lup,
            Tg,
            ea,
            esky,
            I0,
            CI,
            shadow,
            firstdaytime,
            timestepdec,
            timeadd,
            Tgmap1,
            Tgmap1E,
            Tgmap1S,
            Tgmap1W,
            Tgmap1N,
            Keast,
            Ksouth,
            Kwest,
            Knorth,
            Least,
            Lsouth,
            Lwest,
            Lnorth,
            KsideI,
            TgOut1,
            TgOut,
            radIout,
            radDout,
            Lside,
            Lsky_patch_characteristics,
            CI_Tg,
            CI_TgG,
            KsideD,
            dRad,
            Kside,
        ) = so.Solweig_2022a_calc(
            i,
            dsm,
            dsm_scale,
            dsm_height,
            dsm_width,
            svf,
            svfN,
            svfW,
            svfE,
            svfS,
            svfveg,
            svfNveg,
            svfEveg,
            svfSveg,
            svfWveg,
            svfaveg,
            svfEaveg,
            svfSaveg,
            svfWaveg,
            svfNaveg,
            veg_dsm,
            veg_dsm_2,
            albedo_bldg,
            body_shortwave_absorp,
            body_longwave_absorp,
            emmisiv_bldg,
            Fside,
            Fup,
            Fcyl,
            altitude[0][i],
            azimuth[0][i],
            zen[0][i],
            jday[0][i],
            usevegdem,
            int(estimate_radiation_from_global),
            buildings,
            location,
            psi[0][i],
            landcover,  # set to 0
            None,  # lcgrid
            dectime[i],
            altmax[0][i],
            wa_rast,
            wh_rast,
            as_cylinder,
            0,  # elvis
            Ta[i],
            RH[i],
            radG[i],
            radD[i],
            radI[i],
            P[i],
            amaxvalue,
            bush,
            None,  # Twater - used for landcover
            TgK,
            Tstart,
            alb_grid,
            emis_grid,
            TgK_wall,
            Tstart_wall,
            TmaxLST,
            TmaxLST_wall,
            first,
            second,
            svfalfa,
            svfbuveg,
            firstdaytime,
            timeadd,
            timestepdec,
            Tgmap1,
            Tgmap1E,
            Tgmap1S,
            Tgmap1W,
            Tgmap1N,
            CI,
            TgOut1,
            diffsh,
            shmat,
            vegshmat,
            vbshvegshmat,
            anisotropic_sky,
            asvf,
            patch_option,
        )

        if i < first_unique_day.shape[0]:
            I0_array[i] = I0

        tmrtplot = tmrtplot + Tmrt

        if altitude[0][i] > 0:
            w = "D"
        else:
            w = "N"

        if hours[i] < 10:
            XH = "0"
        else:
            XH = ""

        if minu[i] < 10:
            XM = "0"
        else:
            XM = ""

        time_code = (
            str(int(YYYY[0, i])) + "_" + str(int(DOY[i])) + "_" + XH + str(int(hours[i])) + XM + str(int(minu[i])) + w
        )

        if pois_gdf is not None:
            for idx, row in pois_gdf.iterrows():
                centroid = row["geometry"].centroid
                row_idx, col_idx = rowcol(dsm_trf_affine, centroid.x, centroid.y)
                row_idx = int(row_idx)
                col_idx = int(col_idx)
                result_row = {
                    "geometry": row["geometry"],
                    "poi_idx": idx,
                    "yyyy": YYYY[0, i],
                    "id": jday[0, i],
                    "it": hours[i],
                    "imin": minu[i],
                    "dectime": dectime[i],
                    "altitude": altitude[0, i],
                    "azimuth": azimuth[0, i],
                    "kdir": radIout,
                    "kdiff": radDout,
                    "kglobal": radG[i],
                    "kdown": Kdown[row_idx, col_idx],
                    "kup": Kup[row_idx, col_idx],
                    "keast": Keast[row_idx, col_idx],
                    "ksouth": Ksouth[row_idx, col_idx],
                    "kwest": Kwest[row_idx, col_idx],
                    "knorth": Knorth[row_idx, col_idx],
                    "ldown": Ldown[row_idx, col_idx],
                    "lup": Lup[row_idx, col_idx],
                    "least": Least[row_idx, col_idx],
                    "lsouth": Lsouth[row_idx, col_idx],
                    "lwest": Lwest[row_idx, col_idx],
                    "lnorth": Lnorth[row_idx, col_idx],
                    "Ta": Ta[i],
                    "Tg": TgOut[row_idx, col_idx],
                    "RH": RH[i],
                    "Esky": esky,
                    "Tmrt": Tmrt[row_idx, col_idx],
                    "I0": I0,
                    "CI": CI,
                    "Shadow": shadow[row_idx, col_idx],
                    "SVF_b": svf[row_idx, col_idx],
                    "SVF_bv": svfbuveg[row_idx, col_idx],
                    "KsideI": KsideI[row_idx, col_idx],
                }
                # Recalculating wind speed based on powerlaw
                WsPET = (1.1 / wind_sensor_ht) ** 0.2 * Ws[i]
                WsUTCI = (10.0 / wind_sensor_ht) ** 0.2 * Ws[i]
                resultPET = p._PET(
                    Ta[i],
                    RH[i],
                    Tmrt[row_idx, col_idx],
                    WsPET,
                    weight,
                    age,
                    pers_ht,
                    activity,
                    clothing,
                    sex,
                )
                result_row["PET"] = resultPET
                resultUTCI = utci.utci_calculator(Ta[i], RH[i], Tmrt[row_idx, col_idx], WsUTCI)
                result_row["UTCI"] = resultUTCI
                result_row["CI_Tg"] = CI_Tg
                result_row["CI_TgG"] = CI_TgG
                result_row["KsideD"] = KsideD[row_idx, col_idx]
                result_row["Lside"] = Lside[row_idx, col_idx]
                result_row["diffDown"] = dRad[row_idx, col_idx]
                result_row["Kside"] = Kside[row_idx, col_idx]
                poi_results.append(result_row)

        common.save_raster(
            out_path_str + "/Tmrt/Tmrt_" + time_code + ".tif", Tmrt, dsm_transf, dsm_crs, coerce_f64_to_f32=True
        )
        common.save_raster(
            out_path_str
            + "/shadows/Shadow_"
            + str(int(YYYY[0, i]))
            + "_"
            + str(int(DOY[i]))
            + "_"
            + XH
            + str(int(hours[i]))
            + XM
            + str(int(minu[i]))
            + w
            + ".tif",
            shadow,
            dsm_transf,
            dsm_crs,
            coerce_f64_to_f32=True,
        )

    # After the main loop, write all POI results to a single GPKG file with multi-index
    if pois_gdf is not None and len(poi_results) > 0:
        poi_df = gpd.GeoDataFrame(poi_results, geometry="geometry", crs=pois_gdf.crs)
        # Create a datetime column for multi-index
        poi_df["snapshot"] = pd.to_datetime(
            poi_df["yyyy"].astype(int).astype(str)
            + "-"
            + poi_df["id"].astype(int).astype(str).str.zfill(3)
            + " "
            + poi_df["it"].astype(int).astype(str).str.zfill(2)
            + ":"
            + poi_df["imin"].astype(int).astype(str).str.zfill(2),
            format="%Y-%j %H:%M",
        )

        # GPD doesn't handle multi-index
        poi_df.to_file(out_path_str + "/POI.gpkg", driver="GPKG")

    # Output I0 vs. Kglobal plot
    radG_for_plot = radG[first_unique_day[0] == DOY]
    hours_for_plot = hours[first_unique_day[0] == DOY]
    fig, ax = plt.subplots()
    ax.plot(hours_for_plot, I0_array, label="I0")
    ax.plot(hours_for_plot, radG_for_plot, label="Kglobal")
    ax.set_ylabel("Shortwave radiation [$Wm^{-2}$]")
    ax.set_xlabel("Hours")
    ax.set_title("UTC" + str(int(utc)))
    ax.legend()
    fig.savefig(out_path_str + "/metCheck.png", dpi=150)

    # Copying met file for SpatialTC
    umep_df.to_csv(out_path_str + "/metforcing.csv")

    tmrtplot = tmrtplot / Ta.__len__()  # fix average Tmrt instead of sum, 20191022
    common.save_raster(out_path_str + "/Tmrt_average.tif", tmrtplot, dsm_transf, dsm_crs, coerce_f64_to_f32=True)

    rmtree(temp_dir, ignore_errors=True)



import numpy as np

def Tgmaps_v1(lc_grid, solweig_parameters):

    #Tgmaps_v1 Populates grids with cooeficients for Tg wave
    #   Detailed explanation goes here
    lc_grid[lc_grid >= 100] = 2
    id = np.unique(lc_grid)
    id = lc_grid[lc_grid <= 7].astype(int)
    TgK = np.copy(lc_grid)
    Tstart = np.copy(lc_grid)
    alb_grid = np.copy(lc_grid)
    emis_grid = np.copy(lc_grid)
    TmaxLST = np.copy(lc_grid)

    for i in id:
        # row = (lc_class[:, 0] == id[i])
        Tstart[Tstart == i] = solweig_parameters['Tstart']['Value'][solweig_parameters['Names']['Value'][str(i)]]
        alb_grid[alb_grid == i] = solweig_parameters['Albedo']['Effective']['Value'][solweig_parameters['Names']['Value'][str(i)]]
        emis_grid[emis_grid == i] = solweig_parameters['Emissivity']['Value'][solweig_parameters['Names']['Value'][str(i)]]
        TmaxLST[TmaxLST == i] = solweig_parameters['TmaxLST']['Value'][solweig_parameters['Names']['Value'][str(i)]]
        TgK[TgK == i] = solweig_parameters['Ts_deg']['Value'][solweig_parameters['Names']['Value'][str(i)]]

    TgK_wall = solweig_parameters['Ts_deg']['Value']['Walls']
    Tstart_wall = solweig_parameters['Tstart']['Value']['Walls']
    TmaxLST_wall = solweig_parameters['TmaxLST']['Value']['Walls']

    return TgK, Tstart, alb_grid, emis_grid, TgK_wall, Tstart_wall, TmaxLST, TmaxLST_wall



import numpy as np
import pyproj
import pytest

from umep.common import load_raster, save_raster, xy_to_lnglat


def _make_gt(width, height, pixel_size=1):
    # top-left origin at (0, height) so bounds are [0, 0, width, height]
    return [0.0, float(pixel_size), 0.0, float(height), 0.0, -float(pixel_size)]


def test_save_and_load_raster_roundtrip(tmp_path):
    out = tmp_path / "out_dir" / "test.tif"
    data = np.arange(25, dtype=np.float32).reshape(5, 5)
    trf = _make_gt(5, 5)
    crs_wkt = pyproj.CRS.from_epsg(4326).to_wkt()

    # save and load
    save_raster(str(out), data, trf, crs_wkt, no_data_val=-9999, coerce_f64_to_f32=True)
    rast, trf_out, crs_out, nodata = load_raster(str(out), coerce_f64_to_f32=True)

    np.testing.assert_array_equal(rast, data)
    assert isinstance(trf_out, list) and len(trf_out) == 6
    assert np.allclose(trf_out, trf)
    assert nodata == -9999 or nodata is None
    # parsed CRS should map to EPSG:4326
    parsed_crs = pyproj.CRS.from_wkt(crs_out) if crs_out is not None else None
    assert parsed_crs is not None and parsed_crs.to_epsg() == 4326


def test_load_raster_with_bbox(tmp_path):
    out = tmp_path / "bbox_dir" / "bbox.tif"
    data = np.arange(100, dtype=np.float32).reshape(10, 10)
    trf = _make_gt(10, 10)
    crs_wkt = pyproj.CRS.from_epsg(4326).to_wkt()
    save_raster(str(out), data, trf, crs_wkt, -9999, coerce_f64_to_f32=True)

    # bbox in spatial coords: [minx, miny, maxx, maxy]
    # For our geotransform bounds = [0,0,10,10]
    bbox = [2, 2, 5, 5]
    rast_crop, trf_crop, crs_crop, nd = load_raster(str(out), bbox=bbox, coerce_f64_to_f32=True)

    # Expected slice computed from implementation mapping
    assert rast_crop.shape == (3, 3)
    expected = data[5:8, 2:5]  # as per transform -> yoff = 10-5 =5, xoff =2
    np.testing.assert_array_equal(rast_crop, expected)
    assert isinstance(trf_crop, list) and len(trf_crop) == 6
    expected_trf = [bbox[0], trf[1], 0.0, bbox[3], 0.0, trf[5]]
    assert np.allclose(trf_crop, expected_trf)
    assert crs_crop is not None and pyproj.CRS.from_wkt(crs_crop) == pyproj.CRS.from_wkt(crs_wkt)
    assert nd == -9999


def test_xy_to_lnglat_scalar_and_array():
    # WGS84 should be identity
    crs_wkt = pyproj.CRS.from_epsg(4326).to_wkt()
    x, y = 10.0, 20.0
    lon, lat = xy_to_lnglat(crs_wkt, x, y)
    assert lon == pytest.approx(10.0)
    assert lat == pytest.approx(20.0)

    # array case
    xa = np.array([0.0, 30.0])
    ya = np.array([0.0, -15.0])
    lons, lats = xy_to_lnglat(crs_wkt, xa, ya)
    assert np.array_equal(lons, np.array([0.0, 30.0]))
    assert np.array_equal(lats, np.array([0.0, -15.0]))



# This is the main function of the SOLWEIG model
# 2025-Mar-21
# Fredrik Lindberg, fredrikl@gvc.gu.se
# Goteborg Urban Climate Group
# Gothenburg University

# sommon imports
import json
import zipfile
from shutil import copyfile

import matplotlib.pylab as plt
import numpy as np
import pandas as pd

from ...functions import wallalgorithms as wa
from ...functions.SOLWEIGpython import PET_calculations as p
from ...functions.SOLWEIGpython import Solweig_2025a_calc_forprocessing as so
from ...functions.SOLWEIGpython import UTCI_calculations as utci
from ...functions.SOLWEIGpython.CirclePlotBar import PolarBarPlot
from ...functions.SOLWEIGpython.patch_characteristics import hemispheric_image
from ...functions.SOLWEIGpython.Tgmaps_v1 import Tgmaps_v1
from ...functions.SOLWEIGpython.wall_surface_temperature import load_walls
from ...functions.SOLWEIGpython.wallsAsNetCDF import walls_as_netcdf
from ...util.SEBESOLWEIGCommonFiles.clearnessindex_2013b import clearnessindex_2013b
from ...util.SEBESOLWEIGCommonFiles.Solweig_v2015_metdata_noload import Solweig_2015a_metdata_noload
from ...util.umep_solweig_export_component import read_solweig_config

# imports from osgeo/qgis dependency
try:
    from osgeo.gdalconst import *

    from ...functions.SOLWEIGpython.wallOfInterest import pointOfInterest
except:
    pass

# imports for standalone
try:
    import geopandas as gpd
    from rasterio.transform import Affine, rowcol
    from tqdm import tqdm

    from umep import common
except:
    pass


def solweig_run(configPath, feedback):
    """
    Input:
    configPath : config file including geodata paths and settings.
    feedback : To communicate with qgis gui. Set to None if standalone
    """

    # Load config file
    configDict = read_solweig_config(configPath)

    # Load parameters settings for SOLWEIG
    with open(configDict["para_json_path"]) as jsn:
        param = json.load(jsn)

    standAlone = int(configDict["standalone"])

    # Load DSM
    dsm_arr, dsm_trf_arr, dsm_crs_wkt, dsm_nd_val = common.load_raster(
        configDict["filepath_dsm"], bbox=None, coerce_f64_to_f32=True
    )
    # trf is a list: [top left x, w-e pixel size, rotation, top left y, rotation, n-s pixel size]
    scale = 1 / dsm_trf_arr[1]  # pixel resolution in metres
    left_x = dsm_trf_arr[0]
    top_y = dsm_trf_arr[3]
    lng, lat = common.xy_to_lnglat(dsm_crs_wkt, left_x, top_y)
    rows = dsm_arr.shape[0]
    cols = dsm_arr.shape[1]

    # response to issue #85
    dsm_arr[dsm_arr == dsm_nd_val] = 0.0
    if dsm_arr.min() < 0:
        dsmraise = np.abs(dsm_arr.min())
        dsm_arr = dsm_arr + dsmraise
    else:
        dsmraise = 0

    alt = np.median(dsm_arr)
    if alt < 0:
        alt = 3

    # Vegetation
    transVeg = param["Tree_settings"]["Value"]["Transmissivity"]
    trunkratio = param["Tree_settings"]["Value"]["Trunk_ratio"]
    usevegdem = int(configDict["usevegdem"])
    if usevegdem == 1:
        vegdsm, _, _, _ = common.load_raster(configDict["filepath_cdsm"], bbox=None, coerce_f64_to_f32=True)
        if configDict["filepath_tdsm"] != "":
            vegdsm2, _, _, _ = common.load_raster(configDict["filepath_tdsm"], bbox=None, coerce_f64_to_f32=True)
        else:
            vegdsm2 = vegdsm * trunkratio
    else:
        vegdsm = 0
        vegdsm2 = 0

    # Land cover
    landcover = int(configDict["landcover"])
    if landcover == 1:
        lcgrid, _, _, _ = common.load_raster(configDict["filepath_lc"], bbox=None, coerce_f64_to_f32=True)
    else:
        lcgrid = 0

    # DEM for buildings #TODO: fix nodata in standalone
    demforbuild = int(configDict["demforbuild"])
    if demforbuild == 1:
        dem, _, _, dem_nd_val = common.load_raster(configDict["filepath_dem"], bbox=None, coerce_f64_to_f32=True)
        # response to issue and #230
        dem[dem == dem_nd_val] = 0.0
        if dem.min() < 0:
            demraise = np.abs(dem.min())
            dem = dem + demraise
        else:
            demraise = 0

    # SVF
    zip = zipfile.ZipFile(configDict["input_svf"], "r")
    zip.extractall(configDict["working_dir"])
    zip.close()

    svf, _, _, _ = common.load_raster(configDict["working_dir"] + "/svf.tif", bbox=None, coerce_f64_to_f32=True)
    svfN, _, _, _ = common.load_raster(configDict["working_dir"] + "/svfN.tif", bbox=None, coerce_f64_to_f32=True)
    svfS, _, _, _ = common.load_raster(configDict["working_dir"] + "/svfS.tif", bbox=None, coerce_f64_to_f32=True)
    svfE, _, _, _ = common.load_raster(configDict["working_dir"] + "/svfE.tif", bbox=None, coerce_f64_to_f32=True)
    svfW, _, _, _ = common.load_raster(configDict["working_dir"] + "/svfW.tif", bbox=None, coerce_f64_to_f32=True)

    if usevegdem == 1:
        svfveg, _, _, _ = common.load_raster(
            configDict["working_dir"] + "/svfveg.tif", bbox=None, coerce_f64_to_f32=True
        )
        svfNveg, _, _, _ = common.load_raster(
            configDict["working_dir"] + "/svfNveg.tif", bbox=None, coerce_f64_to_f32=True
        )
        svfSveg, _, _, _ = common.load_raster(
            configDict["working_dir"] + "/svfSveg.tif", bbox=None, coerce_f64_to_f32=True
        )
        svfEveg, _, _, _ = common.load_raster(
            configDict["working_dir"] + "/svfEveg.tif", bbox=None, coerce_f64_to_f32=True
        )
        svfWveg, _, _, _ = common.load_raster(
            configDict["working_dir"] + "/svfWveg.tif", bbox=None, coerce_f64_to_f32=True
        )

        svfaveg, _, _, _ = common.load_raster(
            configDict["working_dir"] + "/svfaveg.tif", bbox=None, coerce_f64_to_f32=True
        )
        svfNaveg, _, _, _ = common.load_raster(
            configDict["working_dir"] + "/svfNaveg.tif", bbox=None, coerce_f64_to_f32=True
        )
        svfSaveg, _, _, _ = common.load_raster(
            configDict["working_dir"] + "/svfSaveg.tif", bbox=None, coerce_f64_to_f32=True
        )
        svfEaveg, _, _, _ = common.load_raster(
            configDict["working_dir"] + "/svfEaveg.tif", bbox=None, coerce_f64_to_f32=True
        )
        svfWaveg, _, _, _ = common.load_raster(
            configDict["working_dir"] + "/svfWaveg.tif", bbox=None, coerce_f64_to_f32=True
        )
    else:
        svfveg = np.ones((rows, cols))
        svfNveg = np.ones((rows, cols))
        svfSveg = np.ones((rows, cols))
        svfEveg = np.ones((rows, cols))
        svfWveg = np.ones((rows, cols))
        svfaveg = np.ones((rows, cols))
        svfNaveg = np.ones((rows, cols))
        svfSaveg = np.ones((rows, cols))
        svfEaveg = np.ones((rows, cols))
        svfWaveg = np.ones((rows, cols))

    tmp = svf + svfveg - 1.0
    tmp[tmp < 0.0] = 0.0
    # %matlab crazyness around 0
    svfalfa = np.arcsin(np.exp(np.log(1.0 - tmp) / 2.0))

    wallheight, _, _, _ = common.load_raster(configDict["filepath_wh"], bbox=None, coerce_f64_to_f32=True)
    wallaspect, _, _, _ = common.load_raster(configDict["filepath_wa"], bbox=None, coerce_f64_to_f32=True)

    # Metdata
    headernum = 1
    delim = " "
    Twater = []

    metdata = np.loadtxt(configDict["input_met"], skiprows=headernum, delimiter=delim)

    location = {"longitude": lng, "latitude": lat, "altitude": alt}
    YYYY, altitude, azimuth, zen, jday, leafon, dectime, altmax = Solweig_2015a_metdata_noload(
        metdata, location, int(configDict["utc"])
    )

    DOY = metdata[:, 1]
    hours = metdata[:, 2]
    minu = metdata[:, 3]
    Ta = metdata[:, 11]
    RH = metdata[:, 10]
    radG = metdata[:, 14]
    radD = metdata[:, 21]
    radI = metdata[:, 22]
    P = metdata[:, 12]
    Ws = metdata[:, 9]

    # POIs check
    if configDict["poi_file"] != "":  # usePOI:
        header = (
            "yyyy id   it imin dectime altitude azimuth kdir kdiff kglobal kdown   kup    keast ksouth "
            "kwest knorth ldown   lup    least lsouth lwest  lnorth   Ta      Tg     RH    Esky   Tmrt    "
            "I0     CI   Shadow  SVF_b  SVF_bv KsideI PET UTCI  CI_Tg   CI_TgG  KsideD  Lside   diffDown    Kside"
        )
        poi_field = configDict["poi_field"]
        if standAlone == 0:
            poi_field = configDict["poi_field"]
            poisxy, poiname = pointOfInterest(configDict["poi_file"], poi_field, scale, gdal_dsm)
        else:
            pois_gdf = gpd.read_file(configDict["poi_file"])
            numfeat = pois_gdf.shape[0]
            poisxy = np.zeros((numfeat, 3)) - 999
            poiname = []
            for idx, row in pois_gdf.iterrows():
                y, x = rowcol(
                    Affine.from_gdal(*dsm_trf_arr), row["geometry"].centroid.x, row["geometry"].centroid.y
                )  # TODO: This produce different result since no standalone round coordinates
                if configDict["poi_field"]:
                    poiname.append(row[configDict["poi_field"]])
                else:
                    poiname.append(str(idx))
                poisxy[idx, 0] = idx
                poisxy[idx, 1] = x
                poisxy[idx, 2] = y

        for k in range(0, poisxy.shape[0]):
            poi_save = []
            data_out = configDict["output_dir"] + "POI_" + str(poiname[k]) + ".txt"
            np.savetxt(data_out, poi_save, delimiter=" ", header=header, comments="")
        print(poisxy)
        # Num format for POI output
        numformat = "%d %d %d %d %.5f " + "%.2f " * 36

        # Other PET variables
        sensorheight = param["Wind_Height"]["Value"]["magl"]
        age = param["PET_settings"]["Value"]["Age"]
        mbody = param["PET_settings"]["Value"]["Weight"]
        ht = param["PET_settings"]["Value"]["Height"]
        clo = param["PET_settings"]["Value"]["clo"]
        activity = param["PET_settings"]["Value"]["Activity"]
        sex = param["PET_settings"]["Value"]["Sex"]
    else:
        poisxy = None

    # Posture settings
    if param["Tmrt_params"]["Value"]["posture"] == "Standing":
        Fside = param["Posture"]["Standing"]["Value"]["Fside"]
        Fup = param["Posture"]["Standing"]["Value"]["Fup"]
        height = param["Posture"]["Standing"]["Value"]["height"]
        Fcyl = param["Posture"]["Standing"]["Value"]["Fcyl"]
        pos = 1
    else:
        Fside = param["Posture"]["Sitting"]["Value"]["Fside"]
        Fup = param["Posture"]["Sitting"]["Value"]["Fup"]
        height = param["Posture"]["Sitting"]["Value"]["height"]
        Fcyl = param["Posture"]["Sitting"]["Value"]["Fcyl"]
        pos = 0

    # Radiative surface influence, Rule of thumb by Schmid et al. (1990).
    first = np.round(height)
    if first == 0.0:
        first = 1.0
    second = np.round(height * 20.0)

    if usevegdem == 1:
        # Conifer or deciduous
        if configDict["conifer_bool"]:
            leafon = np.ones((1, DOY.shape[0]))
        else:
            leafon = np.zeros((1, DOY.shape[0]))
            if param["Tree_settings"]["Value"]["First_day_leaf"] > param["Tree_settings"]["Value"]["Last_day_leaf"]:
                leaf_bool = (param["Tree_settings"]["Value"]["First_day_leaf"] < DOY) | (
                    param["Tree_settings"]["Value"]["Last_day_leaf"] > DOY
                )
            else:
                leaf_bool = (param["Tree_settings"]["Value"]["First_day_leaf"] < DOY) & (
                    param["Tree_settings"]["Value"]["Last_day_leaf"] > DOY
                )
            leafon[0, leaf_bool] = 1

        # % Vegetation transmittivity of shortwave radiation
        psi = leafon * transVeg
        psi[leafon == 0] = 0.5
        # amaxvalue
        vegmax = vegdsm.max()
        amaxvalue = dsm_arr.max() - dsm_arr.min()
        amaxvalue = np.maximum(amaxvalue, vegmax)

        # Elevation vegdsms if buildingDEM includes ground heights
        vegdsm = vegdsm + dsm_arr
        vegdsm[vegdsm == dsm_arr] = 0
        vegdsm2 = vegdsm2 + dsm_arr
        vegdsm2[vegdsm2 == dsm_arr] = 0

        # % Bush separation
        bush = np.logical_not(vegdsm2 * vegdsm) * vegdsm

        svfbuveg = svf - (1.0 - svfveg) * (1.0 - transVeg)  # % major bug fixed 20141203
    else:
        psi = leafon * 0.0 + 1.0
        svfbuveg = svf
        bush = np.zeros([rows, cols])
        amaxvalue = 0

    # Initialization of maps
    Knight = np.zeros((rows, cols))
    Tgmap1 = np.zeros((rows, cols))
    Tgmap1E = np.zeros((rows, cols))
    Tgmap1S = np.zeros((rows, cols))
    Tgmap1W = np.zeros((rows, cols))
    Tgmap1N = np.zeros((rows, cols))

    # Create building boolean raster from either land cover or height rasters
    if demforbuild == 0:
        buildings = np.copy(lcgrid)
        buildings[buildings == 7] = 1
        buildings[buildings == 6] = 1
        buildings[buildings == 5] = 1
        buildings[buildings == 4] = 1
        buildings[buildings == 3] = 1
        buildings[buildings == 2] = 0
    else:
        buildings = dsm_arr - dem
        buildings[buildings < 2.0] = 1.0
        buildings[buildings >= 2.0] = 0.0

    if int(configDict["savebuild"]) == 1:
        common.save_raster(
            configDict["output_dir"] + "/buildings.tif",
            buildings,
            dsm_trf_arr,
            dsm_crs_wkt,
            dsm_nd_val,
            coerce_f64_to_f32=True,
        )

    # Import shadow matrices (Anisotropic sky)
    anisotropic_sky = int(configDict["aniso"])
    if anisotropic_sky == 1:  # UseAniso
        data = np.load(configDict["input_aniso"])
        shmat = data["shadowmat"]
        vegshmat = data["vegshadowmat"]
        vbshvegshmat = data["vbshmat"]
        if usevegdem == 1:
            diffsh = np.zeros((rows, cols, shmat.shape[2]))
            for i in range(0, shmat.shape[2]):
                diffsh[:, :, i] = shmat[:, :, i] - (1 - vegshmat[:, :, i]) * (
                    1 - transVeg
                )  # changes in psi not implemented yet
        else:
            diffsh = shmat

        # Estimate number of patches based on shadow matrices
        if shmat.shape[2] == 145:
            patch_option = 1  # patch_option = 1 # 145 patches
        elif shmat.shape[2] == 153:
            patch_option = 2  # patch_option = 2 # 153 patches
        elif shmat.shape[2] == 306:
            patch_option = 3  # patch_option = 3 # 306 patches
        elif shmat.shape[2] == 612:
            patch_option = 4  # patch_option = 4 # 612 patches

        # asvf to calculate sunlit and shaded patches
        asvf = np.arccos(np.sqrt(svf))

        # Empty array for steradians
        steradians = np.zeros(shmat.shape[2])
    else:
        # anisotropic_sky = 0
        diffsh = None
        shmat = None
        vegshmat = None
        vbshvegshmat = None
        asvf = None
        patch_option = 0
        steradians = 0

    # % Ts parameterisation maps
    if landcover == 1.0:
        # Get land cover properties for Tg wave (land cover scheme based on Bogren et al. 2000, explained in Lindberg et al., 2008 and Lindberg, Onomura & Grimmond, 2016)
        [TgK, Tstart, alb_grid, emis_grid, TgK_wall, Tstart_wall, TmaxLST, TmaxLST_wall] = Tgmaps_v1(
            lcgrid.copy(), param
        )
    else:
        TgK = Knight + param["Ts_deg"]["Value"]["Cobble_stone_2014a"]
        Tstart = Knight - param["Tstart"]["Value"]["Cobble_stone_2014a"]
        TmaxLST = param["TmaxLST"]["Value"]["Cobble_stone_2014a"]
        alb_grid = Knight + param["Albedo"]["Effective"]["Value"]["Cobble_stone_2014a"]
        emis_grid = Knight + param["Emissivity"]["Value"]["Cobble_stone_2014a"]
        TgK_wall = param["Ts_deg"]["Value"]["Walls"]
        Tstart_wall = param["Tstart"]["Value"]["Walls"]
        TmaxLST_wall = param["TmaxLST"]["Value"]["Walls"]

    # Import data for wall temperature parameterization TODO: fix for standalone
    wallScheme = int(configDict["wallscheme"])
    if wallScheme == 1:
        wallData = np.load(configDict["input_wall"])
        voxelMaps = wallData["voxelId"]
        voxelTable = wallData["voxelTable"]
        # Get wall type from standalone
        if standAlone == 1:
            wall_type_standalone = {"Brick_wall": "100", "Concrete_wall": "101", "Wood_wall": "102"}
            wall_type = wall_type_standalone[configDict["walltype"]]
        else:
            # Get wall type set in GUI
            wall_type = configDict[
                "walltype"
            ]  # str(100 + int(self.parameterAsString(parameters, self.WALL_TYPE, context))) #TODO

        # Get heights of walls including corners
        walls_scheme = wa.findwalls_sp(dsm_arr, 2, np.array([[1, 1, 1], [1, 0, 1], [1, 1, 1]]))
        # Get aspects of walls including corners
        dirwalls_scheme = wa.filter1Goodwin_as_aspect_v3(walls_scheme.copy(), scale, dsm_arr, None, 100.0 / 180.0)

        # Used in wall temperature parameterization scheme
        first_timestep = (
            pd.to_datetime(YYYY[0][0], format="%Y")
            + pd.to_timedelta(DOY[0] - 1, unit="d")
            + pd.to_timedelta(hours[0], unit="h")
            + pd.to_timedelta(minu[0], unit="m")
        )
        second_timestep = (
            pd.to_datetime(YYYY[0][1], format="%Y")
            + pd.to_timedelta(DOY[1] - 1, unit="d")
            + pd.to_timedelta(hours[1], unit="h")
            + pd.to_timedelta(minu[1], unit="m")
        )

        timeStep = (second_timestep - first_timestep).seconds

        # Load voxelTable as Pandas DataFrame
        voxelTable, dirwalls_scheme = load_walls(
            voxelTable, param, wall_type, dirwalls_scheme, Ta[0], timeStep, alb_grid, landcover, lcgrid, dsm_arr
        )

        # Use wall of interest
        woi_file = configDict["woi_file"]
        if woi_file:
            # (dsm_minx, dsm_x_size, dsm_x_rotation, dsm_miny, dsm_y_rotation, dsm_y_size) = gdal_dsm.GetGeoTransform() #TODO: fix for standalone
            if standAlone == 0:
                woi_field = configDict["woi_field"]  # self.parameterAsStrings(parameters, self.WOI_FIELD, context)
                woisxy, woiname = pointOfInterest(configDict["woi_file"], woi_field, scale, gdal_dsm)
            else:
                pois_gdf = gpd.read_file(configDict["poi_file"])
                numfeat = pois_gdf.shape[0]
                poisxy = np.zeros((numfeat, 3)) - 999
                for idx, row in pois_gdf.iterrows():
                    y, x = rowcol(
                        dsm_trf_arr, row["geometry"].centroid.x, row["geometry"].centroid.y
                    )  # TODO: This produce different result since no standalone round coordinates
                    poiname.append(row[configDict["poi_field"]])
                    poisxy[idx, 0] = idx
                    poisxy[idx, 1] = x
                    poisxy[idx, 2] = y

        # Create pandas datetime object to be used when createing an xarray DataSet where wall temperatures/radiation is stored and eventually saved as a NetCDf
        if configDict["wallnetcdf"] == 1:
            met_for_xarray = (
                pd.to_datetime(YYYY[0][:], format="%Y")
                + pd.to_timedelta(DOY - 1, unit="d")
                + pd.to_timedelta(hours, unit="h")
                + pd.to_timedelta(minu, unit="m")
            )
    else:
        wallScheme = 0
        voxelMaps = 0
        voxelTable = 0
        timeStep = 0
        # thermal_effusivity = 0
        walls_scheme = np.ones((rows, cols)) * 10.0
        dirwalls_scheme = np.ones((rows, cols)) * 10.0

    # Initialisation of time related variables
    if Ta.__len__() == 1:
        timestepdec = 0
    else:
        timestepdec = dectime[1] - dectime[0]
    timeadd = 0.0
    firstdaytime = 1.0

    # Save hemispheric image
    if anisotropic_sky == 1:
        if poisxy is not None:
            patch_characteristics = hemispheric_image(poisxy, shmat, vegshmat, vbshvegshmat, voxelMaps, wallScheme)

    # If metfile starts at night
    CI = 1.0

    # reading variables from config and parameters that is not yet presented
    cyl = int(configDict["cyl"])
    albedo_b = param["Albedo"]["Effective"]["Value"]["Walls"]
    ewall = param["Emissivity"]["Value"]["Walls"]
    onlyglobal = int(configDict["onlyglobal"])
    elvis = 0.0
    absK = param["Tmrt_params"]["Value"]["absK"]
    absL = param["Tmrt_params"]["Value"]["absL"]

    # Main loop
    tmrtplot = np.zeros((rows, cols))
    TgOut1 = np.zeros((rows, cols))

    # Initiate array for I0 values
    if np.unique(DOY).shape[0] > 1:
        unique_days = np.unique(DOY)
        first_unique_day = DOY[unique_days[0] == DOY]
        I0_array = np.zeros(first_unique_day.shape[0])
    else:
        first_unique_day = DOY.copy()
        I0_array = np.zeros(DOY.shape[0])

    if standAlone == 1:
        progress = tqdm(total=Ta.__len__())

    for i in np.arange(0, Ta.__len__()):
        if feedback is not None:
            feedback.setProgress(int(i * (100.0 / Ta.__len__())))  # move progressbar forward
            if feedback.isCanceled():
                feedback.setProgressText("Calculation cancelled")
                break
        else:
            progress.update(1)

        # Daily water body temperature
        if landcover == 1:
            if (dectime[i] - np.floor(dectime[i])) == 0 or (i == 0):
                Twater = np.mean(Ta[jday[0] == np.floor(dectime[i])])
        # Nocturnal cloudfraction from Offerle et al. 2003
        if (dectime[i] - np.floor(dectime[i])) == 0:
            daylines = np.where(np.floor(dectime) == dectime[i])
            if daylines.__len__() > 1:
                alt = altitude[0][daylines]
                alt2 = np.where(alt > 1)
                rise = alt2[0][0]
                [_, CI, _, _, _] = clearnessindex_2013b(
                    zen[0, i + rise + 1],
                    jday[0, i + rise + 1],
                    Ta[i + rise + 1],
                    RH[i + rise + 1] / 100.0,
                    radG[i + rise + 1],
                    location,
                    P[i + rise + 1],
                )
                if (CI > 1.0) or (np.inf == CI):
                    CI = 1.0
            else:
                CI = 1.0

        (
            Tmrt,
            Kdown,
            Kup,
            Ldown,
            Lup,
            Tg,
            ea,
            esky,
            I0,
            CI,
            shadow,
            firstdaytime,
            timestepdec,
            timeadd,
            Tgmap1,
            Tgmap1E,
            Tgmap1S,
            Tgmap1W,
            Tgmap1N,
            Keast,
            Ksouth,
            Kwest,
            Knorth,
            Least,
            Lsouth,
            Lwest,
            Lnorth,
            KsideI,
            TgOut1,
            TgOut,
            radIout,
            radDout,
            Lside,
            Lsky_patch_characteristics,
            CI_Tg,
            CI_TgG,
            KsideD,
            dRad,
            Kside,
            steradians,
            voxelTable,
        ) = so.Solweig_2025a_calc(
            i,
            dsm_arr,
            scale,
            rows,
            cols,
            svf,
            svfN,
            svfW,
            svfE,
            svfS,
            svfveg,
            svfNveg,
            svfEveg,
            svfSveg,
            svfWveg,
            svfaveg,
            svfEaveg,
            svfSaveg,
            svfWaveg,
            svfNaveg,
            vegdsm,
            vegdsm2,
            albedo_b,
            absK,
            absL,
            ewall,
            Fside,
            Fup,
            Fcyl,
            altitude[0][i],
            azimuth[0][i],
            zen[0][i],
            jday[0][i],
            usevegdem,
            onlyglobal,
            buildings,
            location,
            psi[0][i],
            landcover,
            lcgrid,
            dectime[i],
            altmax[0][i],
            wallaspect,
            wallheight,
            cyl,
            elvis,
            Ta[i],
            RH[i],
            radG[i],
            radD[i],
            radI[i],
            P[i],
            amaxvalue,
            bush,
            Twater,
            TgK,
            Tstart,
            alb_grid,
            emis_grid,
            TgK_wall,
            Tstart_wall,
            TmaxLST,
            TmaxLST_wall,
            first,
            second,
            svfalfa,
            svfbuveg,
            firstdaytime,
            timeadd,
            timestepdec,
            Tgmap1,
            Tgmap1E,
            Tgmap1S,
            Tgmap1W,
            Tgmap1N,
            CI,
            TgOut1,
            diffsh,
            shmat,
            vegshmat,
            vbshvegshmat,
            anisotropic_sky,
            asvf,
            patch_option,
            voxelMaps,
            voxelTable,
            Ws[i],
            wallScheme,
            timeStep,
            steradians,
            walls_scheme,
            dirwalls_scheme,
        )

        # Save I0 for I0 vs. Kdown output plot to check if UTC is off
        if i < first_unique_day.shape[0]:
            I0_array[i] = I0
        elif i == first_unique_day.shape[0]:
            # Output I0 vs. Kglobal plot
            radG_for_plot = radG[first_unique_day[0] == DOY]
            dectime_for_plot = dectime[first_unique_day[0] == DOY]
            fig, ax = plt.subplots()
            ax.plot(dectime_for_plot, I0_array, label="I0")
            ax.plot(dectime_for_plot, radG_for_plot, label="Kglobal")
            ax.set_ylabel("Shortwave radiation [$Wm^{-2}$]")
            ax.set_xlabel("Decimal time")
            ax.set_title("UTC" + str(configDict["utc"]))
            ax.legend()
            fig.savefig(configDict["output_dir"] + "/metCheck.png", dpi=150)

        tmrtplot = tmrtplot + Tmrt

        if altitude[0][i] > 0:
            w = "D"
        else:
            w = "N"

        # Write to POIs
        if poisxy is not None:
            for k in range(0, poisxy.shape[0]):
                poi_save = np.zeros((1, 41))
                poi_save[0, 0] = YYYY[0][i]
                poi_save[0, 1] = jday[0][i]
                poi_save[0, 2] = hours[i]
                poi_save[0, 3] = minu[i]
                poi_save[0, 4] = dectime[i]
                poi_save[0, 5] = altitude[0][i]
                poi_save[0, 6] = azimuth[0][i]
                poi_save[0, 7] = radIout
                poi_save[0, 8] = radDout
                poi_save[0, 9] = radG[i]
                poi_save[0, 10] = Kdown[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 11] = Kup[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 12] = Keast[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 13] = Ksouth[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 14] = Kwest[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 15] = Knorth[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 16] = Ldown[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 17] = Lup[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 18] = Least[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 19] = Lsouth[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 20] = Lwest[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 21] = Lnorth[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 22] = Ta[i]
                poi_save[0, 23] = TgOut[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 24] = RH[i]
                poi_save[0, 25] = esky
                poi_save[0, 26] = Tmrt[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 27] = I0
                poi_save[0, 28] = CI
                poi_save[0, 29] = shadow[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 30] = svf[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 31] = svfbuveg[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 32] = KsideI[int(poisxy[k, 2]), int(poisxy[k, 1])]
                # Recalculating wind speed based on powerlaw
                WsPET = (1.1 / sensorheight) ** 0.2 * Ws[i]
                WsUTCI = (10.0 / sensorheight) ** 0.2 * Ws[i]
                resultPET = p._PET(
                    Ta[i], RH[i], Tmrt[int(poisxy[k, 2]), int(poisxy[k, 1])], WsPET, mbody, age, ht, activity, clo, sex
                )
                poi_save[0, 33] = resultPET
                resultUTCI = utci.utci_calculator(Ta[i], RH[i], Tmrt[int(poisxy[k, 2]), int(poisxy[k, 1])], WsUTCI)
                poi_save[0, 34] = resultUTCI
                poi_save[0, 35] = CI_Tg
                poi_save[0, 36] = CI_TgG
                poi_save[0, 37] = KsideD[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 38] = Lside[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 39] = dRad[int(poisxy[k, 2]), int(poisxy[k, 1])]
                poi_save[0, 40] = Kside[int(poisxy[k, 2]), int(poisxy[k, 1])]
                data_out = configDict["output_dir"] + "/POI_" + str(poiname[k]) + ".txt"
                # f_handle = file(data_out, 'a')
                f_handle = open(data_out, "ab")
                np.savetxt(f_handle, poi_save, fmt=numformat)
                f_handle.close()

        # If wall temperature parameterization scheme is in use
        if configDict["wallscheme"] == 1:  # folderWallScheme: TODO: Fix for standalone
            # Store wall data for output
            if woisxy is not None:
                for k in range(0, woisxy.shape[0]):
                    temp_wall = voxelTable.loc[
                        ((voxelTable["ypos"] == woisxy[k, 2]) & (voxelTable["xpos"] == woisxy[k, 1])), "wallTemperature"
                    ].to_numpy()
                    K_in = voxelTable.loc[
                        ((voxelTable["ypos"] == woisxy[k, 2]) & (voxelTable["xpos"] == woisxy[k, 1])), "K_in"
                    ].to_numpy()
                    L_in = voxelTable.loc[
                        ((voxelTable["ypos"] == woisxy[k, 2]) & (voxelTable["xpos"] == woisxy[k, 1])), "L_in"
                    ].to_numpy()
                    wallShade = voxelTable.loc[
                        ((voxelTable["ypos"] == woisxy[k, 2]) & (voxelTable["xpos"] == woisxy[k, 1])), "wallShade"
                    ].to_numpy()
                    temp_all = np.concatenate([temp_wall, K_in, L_in, wallShade])
                    wall_data = np.zeros((1, 7 + temp_all.shape[0]))
                    # Part of file name (wallid), i.e. WOI_wallid.txt
                    data_out = configDict["output_dir"] + "/WOI_" + str(woiname[k]) + ".txt"
                    if i == 0:
                        # Output file header
                        header = (
                            "yyyy id   it imin dectime Ta  SVF"
                            + " Ts" * temp_wall.shape[0]
                            + " Kin" * K_in.shape[0]
                            + " Lin" * L_in.shape[0]
                            + " shade" * wallShade.shape[0]
                        )
                        woi_save = []  #
                        np.savetxt(data_out, woi_save, delimiter=" ", header=header, comments="")
                    # Fill wall_data with variables
                    wall_data[0, 0] = YYYY[0][i]
                    wall_data[0, 1] = jday[0][i]
                    wall_data[0, 2] = hours[i]
                    wall_data[0, 3] = minu[i]
                    wall_data[0, 4] = dectime[i]
                    wall_data[0, 5] = Ta[i]
                    wall_data[0, 6] = svf[int(woisxy[k, 2]), int(woisxy[k, 1])]
                    wall_data[0, 7:] = temp_all

                    # Num format for output file data
                    woi_numformat = "%d %d %d %d %.5f %.2f %.2f" + " %.2f" * temp_all.shape[0]
                    # Open file, add data, save
                    f_handle = open(data_out, "ab")
                    np.savetxt(f_handle, wall_data, fmt=woi_numformat)
                    f_handle.close()

            # Save wall temperature/radiation as NetCDF TODO: fix for standAlone?
            if configDict["wallnetcdf"] == "1":  # wallNetCDF:
                netcdf_output = configDict["outputDir"] + "/walls.nc"
                walls_as_netcdf(
                    voxelTable, rows, cols, met_for_xarray, i, dsm_arr, configDict["filepath_dsm"], netcdf_output
                )

        if hours[i] < 10:
            XH = "0"
        else:
            XH = ""
        if minu[i] < 10:
            XM = "0"
        else:
            XM = ""

        time_code = (
            str(int(YYYY[0, i])) + "_" + str(int(DOY[i])) + "_" + XH + str(int(hours[i])) + XM + str(int(minu[i])) + w
        )

        if configDict["outputtmrt"] == "1":
            common.save_raster(
                configDict["output_dir"] + "/Tmrt_" + time_code + ".tif",
                Tmrt,
                dsm_trf_arr,
                dsm_crs_wkt,
                dsm_nd_val,
                coerce_f64_to_f32=True,
            )
        if configDict["outputkup"] == "1":
            common.save_raster(
                configDict["output_dir"] + "/Kup_" + time_code + ".tif",
                Kup,
                dsm_trf_arr,
                dsm_crs_wkt,
                dsm_nd_val,
                coerce_f64_to_f32=True,
            )
        if configDict["outputkdown"] == "1":
            common.save_raster(
                configDict["output_dir"] + "/Kdown_" + time_code + ".tif",
                Kdown,
                dsm_trf_arr,
                dsm_crs_wkt,
                dsm_nd_val,
                coerce_f64_to_f32=True,
            )
        if configDict["outputlup"] == "1":
            common.save_raster(
                configDict["output_dir"] + "/Lup_" + time_code + ".tif",
                Lup,
                dsm_trf_arr,
                dsm_crs_wkt,
                dsm_nd_val,
                coerce_f64_to_f32=True,
            )
        if configDict["outputldown"] == "1":
            common.save_raster(
                configDict["output_dir"] + "/Ldown_" + time_code + ".tif",
                Ldown,
                dsm_trf_arr,
                dsm_crs_wkt,
                dsm_nd_val,
                coerce_f64_to_f32=True,
            )
        if configDict["outputsh"] == "1":
            common.save_raster(
                configDict["output_dir"] + "/Shadow_" + time_code + ".tif",
                shadow,
                dsm_trf_arr,
                dsm_crs_wkt,
                dsm_nd_val,
                coerce_f64_to_f32=True,
            )
        if configDict["outputkdiff"] == "1":
            common.save_raster(
                configDict["output_dir"] + "/Kdiff_" + time_code + ".tif",
                dRad,
                dsm_trf_arr,
                dsm_crs_wkt,
                dsm_nd_val,
                coerce_f64_to_f32=True,
            )

        # Sky view image of patches
        if (anisotropic_sky == 1) & (i == 0) & (poisxy is not None):
            for k in range(poisxy.shape[0]):
                Lsky_patch_characteristics[:, 2] = patch_characteristics[:, k]
                skyviewimage_out = configDict["output_dir"] + "/POI_" + str(poiname[k]) + ".png"
                PolarBarPlot(
                    Lsky_patch_characteristics,
                    altitude[0][i],
                    azimuth[0][i],
                    "Hemisphere partitioning",
                    skyviewimage_out,
                    0,
                    5,
                    0,
                )

    # Save files for Tree Planter
    if configDict["outputtreeplanter"] == "1":  # outputTreeplanter:
        if feedback is not None:
            feedback.setProgressText("Saving files for Tree Planter tool")
        # Save DSM
        copyfile(configDict["filepath_dsm"], configDict["output_dir"] + "/DSM.tif")

        # Save CDSM
        if usevegdem == 1:
            copyfile(configDict["filepath_cdsm"], configDict["output_dir"] + "/CDSM.tif")

        albedo_g = param["Albedo"]["Effective"]["Value"]["Cobble_stone_2014a"]
        eground = param["Emissivity"]["Value"]["Cobble_stone_2014a"]

        # Saving settings from SOLWEIG for SOLWEIG1D in TreePlanter
        settingsHeader = "UTC, posture, onlyglobal, landcover, anisotropic, cylinder, albedo_walls, albedo_ground, emissivity_walls, emissivity_ground, absK, absL, elevation, patch_option"
        settingsFmt = (
            "%i",
            "%i",
            "%i",
            "%i",
            "%i",
            "%i",
            "%1.2f",
            "%1.2f",
            "%1.2f",
            "%1.2f",
            "%1.2f",
            "%1.2f",
            "%1.2f",
            "%i",
        )
        settingsData = np.array(
            [
                [
                    int(configDict["utc"]),
                    pos,
                    onlyglobal,
                    landcover,
                    anisotropic_sky,
                    cyl,
                    albedo_b,
                    albedo_g,
                    ewall,
                    eground,
                    absK,
                    absL,
                    alt,
                    patch_option,
                ]
            ]
        )
        np.savetxt(
            configDict["output_dir"] + "/treeplantersettings.txt",
            settingsData,
            fmt=settingsFmt,
            header=settingsHeader,
            delimiter=" ",
        )

    # Copying met file for SpatialTC
    copyfile(configDict["input_met"], configDict["output_dir"] + "/metforcing.txt")

    tmrtplot = tmrtplot / Ta.__len__()  # fix average Tmrt instead of sum, 20191022
    common.save_raster(
        configDict["output_dir"] + "/Tmrt_average.tif",
        tmrtplot,
        dsm_trf_arr,
        dsm_crs_wkt,
        dsm_nd_val,
        coerce_f64_to_f32=True,
    )



# Generated with SMOP  0.41
from libsmop import *
# atm_P.m

    
@function
def atm_P(A=None,*args,**kwargs):
    varargin = atm_P.varargin
    nargin = atm_P.nargin

    # Written by Jenni Vanos, Aug 2009 for use in COMFA CNR to find Rabs.
    
    Po=101.3
# atm_P.m:6
    
    Pa=multiply(Po,exp(- A / 8200))
# atm_P.m:9
    
    #another equation that gives similar results found from website below:
    
    #Pa=Po.*(1-.0000225577.*A).^5.25588;
    
    # A = altitude, units don't matter because it cancels with Ao (which is 0 at surface). 
#Therefore, it will be exp^-1, and Po is surface pressure, which is 1bar, or 100kPa.
    
    # equation from: http://www.engineeringtoolbox.com/air-altitude-pressure-d_462.html


import numpy as np

''' This function calculates whether a point is sunlit or shaded
    based on a sky view factor (in a cylinder), solar altitude, solar azimuth '''

def shaded_or_sunlit(solar_altitude, solar_azimuth, patch_altitude, patch_azimuth, asvf):

    # Patch azimuth in relation to sun azimuth
    patch_to_sun_azi = np.abs(solar_azimuth - patch_azimuth)

    # Degrees to radians
    deg2rad = np.pi/180
    
    # Radians to degrees
    rad2deg = 180/np.pi

    # 
    xi = np.cos(patch_to_sun_azi * deg2rad)

    # 
    yi = 2 * xi * np.tan(solar_altitude * deg2rad)

    hsvf = np.tan(asvf)

    if yi > 0:
        yi_ = 0
    else:
        yi_ = yi

    # 
    tan_delta = hsvf + yi_

    # Degrees where below is in shade and above is sunlit
    sunlit_degrees = np.arctan(tan_delta) * rad2deg

    # Boolean for pixels where patch is sunlit
    sunlit_patches = sunlit_degrees < patch_altitude
    # Boolean for pixels where patch is shaded
    shaded_patches = sunlit_degrees > patch_altitude

    return sunlit_patches, shaded_patches


from __future__ import division
import numpy as np

def diffusefraction(radG,altitude,Kt,Ta,RH):
    """
    This function estimates diffuse and directbeam radiation according to
    Reindl et al (1990), Solar Energy 45:1

    :param radG:
    :param altitude:
    :param Kt: # radiation at the top of the atmosphere
    :param Ta:
    :param RH:
    :return:
    """

    alfa = altitude*(np.pi/180)

    if Ta <= -999.00 or RH <= -999.00 or np.isnan(Ta) or np.isnan(RH):
        if Kt <= 0.3:
            radD = radG*(1.020-0.248*Kt)
        elif Kt > 0.3 and Kt < 0.78:
            radD = radG*(1.45-1.67*Kt)
        else:
            radD = radG*0.147
    else:
        RH = RH/100
        if Kt <= 0.3:
            radD = radG*(1 - 0.232 * Kt + 0.0239 * np.sin(alfa) - 0.000682 * Ta + 0.0195 * RH)
        elif Kt > 0.3 and Kt < 0.78:
            radD = radG*(1.329- 1.716 * Kt + 0.267 * np.sin(alfa) - 0.00357 * Ta + 0.106 * RH)
        else:
            radD = radG*(0.426 * Kt - 0.256 * np.sin(alfa) + 0.00349 * Ta + 0.0734 * RH)

    radI = (radG - radD)/(np.sin(alfa))

    # Corrections for low sun altitudes (20130307)
    if radI < 0:
        radI = 0

    if altitude < 1 and radI > radG:
        radI=radG

    if radD > radG:
        radD = radG

    return radI, radD



from __future__ import absolute_import
import numpy as np
from .Lvikt_veg import Lvikt_veg

def Lside_veg_v2022a(svfS,svfW,svfN,svfE,svfEveg,svfSveg,svfWveg,svfNveg,svfEaveg,svfSaveg,svfWaveg,svfNaveg,azimuth,altitude,Ta,Tw,SBC,ewall,Ldown,esky,t,F_sh,CI,LupE,LupS,LupW,LupN,anisotropic_longwave):

    # This m-file is the current one that estimates L from the four cardinal points 20100414
    
    #Building height angle from svf
    svfalfaE=np.arcsin(np.exp((np.log(1-svfE))/2))
    svfalfaS=np.arcsin(np.exp((np.log(1-svfS))/2))
    svfalfaW=np.arcsin(np.exp((np.log(1-svfW))/2))
    svfalfaN=np.arcsin(np.exp((np.log(1-svfN))/2))
    
    vikttot=4.4897
    aziW=azimuth+t
    aziN=azimuth-90+t
    aziE=azimuth-180+t
    aziS=azimuth-270+t
    
    F_sh = 2*F_sh-1  #(cylindric_wedge scaled 0-1)
    
    c=1-CI
    Lsky_allsky = esky*SBC*((Ta+273.15)**4)*(1-c)+c*SBC*((Ta+273.15)**4)
    
    ## Least
    [viktveg, viktwall, viktsky, viktrefl] = Lvikt_veg(svfE, svfEveg, svfEaveg, vikttot)
    
    if altitude > 0:  # daytime
        alfaB=np.arctan(svfalfaE)
        betaB=np.arctan(np.tan((svfalfaE)*F_sh))
        betasun=((alfaB-betaB)/2)+betaB
        # betasun = np.arctan(0.5*np.tan(svfalfaE)*(1+F_sh)) #TODO This should be considered in future versions
        if (azimuth > (180-t))  and  (azimuth <= (360-t)):
            Lwallsun=SBC*ewall*((Ta+273.15+Tw*np.sin(aziE*(np.pi/180)))**4)*\
                viktwall*(1-F_sh)*np.cos(betasun)*0.5
            Lwallsh=SBC*ewall*((Ta+273.15)**4)*viktwall*F_sh*0.5
        else:
            Lwallsun=0
            Lwallsh=SBC*ewall*((Ta+273.15)**4)*viktwall*0.5
    else: #nighttime
        Lwallsun=0
        Lwallsh=SBC*ewall*((Ta+273.15)**4)*viktwall*0.5
    
    # Longwave from ground (see Lcyl_v2022a for remaining fluxes)
    if anisotropic_longwave == 1:
        Lground=LupE*0.5
        Least=Lground
    else:
        Lsky=((svfE+svfEveg-1)*Lsky_allsky)*viktsky*0.5
        Lveg=SBC*ewall*((Ta+273.15)**4)*viktveg*0.5
        Lground=LupE*0.5
        Lrefl=(Ldown+LupE)*(viktrefl)*(1-ewall)*0.5
        Least=Lsky+Lwallsun+Lwallsh+Lveg+Lground+Lrefl

    # clear alfaB betaB betasun Lsky Lwallsh Lwallsun Lveg Lground Lrefl viktveg viktwall viktsky
    
    ## Lsouth
    [viktveg,viktwall,viktsky,viktrefl]=Lvikt_veg(svfS,svfSveg,svfSaveg,vikttot)
    
    if altitude>0: # daytime
        alfaB=np.arctan(svfalfaS)
        betaB=np.arctan(np.tan((svfalfaS)*F_sh))
        betasun=((alfaB-betaB)/2)+betaB
        # betasun = np.arctan(0.5*np.tan(svfalfaS)*(1+F_sh))
        if (azimuth <= (90-t))  or  (azimuth > (270-t)):
            Lwallsun=SBC*ewall*((Ta+273.15+Tw*np.sin(aziS*(np.pi/180)))**4)*\
                viktwall*(1-F_sh)*np.cos(betasun)*0.5
            Lwallsh=SBC*ewall*((Ta+273.15)**4)*viktwall*F_sh*0.5
        else:
            Lwallsun=0
            Lwallsh=SBC*ewall*((Ta+273.15)**4)*viktwall*0.5
    else: #nighttime
        Lwallsun=0
        Lwallsh=SBC*ewall*((Ta+273.15)**4)*viktwall*0.5

    # Longwave from ground (see Lcyl_v2022a for remaining fluxes)
    if anisotropic_longwave == 1:
        Lground=LupS*0.5
        Lsouth=Lground
    else:
        Lsky=((svfS+svfSveg-1)*Lsky_allsky)*viktsky*0.5
        Lveg=SBC*ewall*((Ta+273.15)**4)*viktveg*0.5
        Lground=LupS*0.5
        Lrefl=(Ldown+LupS)*(viktrefl)*(1-ewall)*0.5
        Lsouth=Lsky+Lwallsun+Lwallsh+Lveg+Lground+Lrefl

    # clear alfaB betaB betasun Lsky Lwallsh Lwallsun Lveg Lground Lrefl viktveg viktwall viktsky
    
    ## Lwest
    [viktveg,viktwall,viktsky,viktrefl]=Lvikt_veg(svfW,svfWveg,svfWaveg,vikttot)
    
    if altitude>0: # daytime
        alfaB=np.arctan(svfalfaW)
        betaB=np.arctan(np.tan((svfalfaW)*F_sh))
        betasun=((alfaB-betaB)/2)+betaB
        # betasun = np.arctan(0.5*np.tan(svfalfaW)*(1+F_sh))
        if (azimuth > (360-t))  or  (azimuth <= (180-t)):
            Lwallsun=SBC*ewall*((Ta+273.15+Tw*np.sin(aziW*(np.pi/180)))**4)*\
                viktwall*(1-F_sh)*np.cos(betasun)*0.5
            Lwallsh=SBC*ewall*((Ta+273.15)**4)*viktwall*F_sh*0.5
        else:
            Lwallsun=0
            Lwallsh=SBC*ewall*((Ta+273.15)**4)*viktwall*0.5
    else: #nighttime
        Lwallsun=0
        Lwallsh=SBC*ewall*((Ta+273.15)**4)*viktwall*0.5

    # Longwave from ground (see Lcyl_v2022a for remaining fluxes)
    if anisotropic_longwave == 1:
        Lground=LupW*0.5
        Lwest=Lground
    else:
        Lsky=((svfW+svfWveg-1)*Lsky_allsky)*viktsky*0.5
        Lveg=SBC*ewall*((Ta+273.15)**4)*viktveg*0.5
        Lground=LupW*0.5
        Lrefl=(Ldown+LupW)*(viktrefl)*(1-ewall)*0.5
        Lwest=Lsky+Lwallsun+Lwallsh+Lveg+Lground+Lrefl

    # clear alfaB betaB betasun Lsky Lwallsh Lwallsun Lveg Lground Lrefl viktveg viktwall viktsky
    
    ## Lnorth
    [viktveg,viktwall,viktsky,viktrefl]=Lvikt_veg(svfN,svfNveg,svfNaveg,vikttot)
    
    if altitude>0: # daytime
        alfaB=np.arctan(svfalfaN)
        betaB=np.arctan(np.tan((svfalfaN)*F_sh))
        betasun=((alfaB-betaB)/2)+betaB
        # betasun = np.arctan(0.5*np.tan(svfalfaN)*(1+F_sh))
        if (azimuth > (90-t))  and  (azimuth <= (270-t)):
            Lwallsun=SBC*ewall*((Ta+273.15+Tw*np.sin(aziN*(np.pi/180)))**4)*\
                viktwall*(1-F_sh)*np.cos(betasun)*0.5
            Lwallsh=SBC*ewall*((Ta+273.15)**4)*viktwall*F_sh*0.5
        else:
            Lwallsun=0
            Lwallsh=SBC*ewall*((Ta+273.15)**4)*viktwall*0.5
    else: #nighttime
        Lwallsun=0
        Lwallsh=SBC*ewall*((Ta+273.15)**4)*viktwall*0.5

    # Longwave from ground (see Lcyl_v2022a for remaining fluxes)
    if anisotropic_longwave == 1:
        Lground=LupN*0.5
        Lnorth=Lground
    else:
        Lsky=((svfN+svfNveg-1)*Lsky_allsky)*viktsky*0.5
        Lveg=SBC*ewall*((Ta+273.15)**4)*viktveg*0.5
        Lground=LupN*0.5
        Lrefl=(Ldown+LupN)*(viktrefl)*(1-ewall)*0.5
        Lnorth=Lsky+Lwallsun+Lwallsh+Lveg+Lground+Lrefl

    # clear alfaB betaB betasun Lsky Lwallsh Lwallsun Lveg Lground Lrefl viktveg viktwall viktsky
    
    return Least, Lsouth, Lwest, Lnorth